//==============================================================================
// File Name: WeightVector.h
// Written by: SJW (14.09.09), following Shai Shalev-Shwartz (28.01.07)
// efficiently implements a weight vector plus intercept, for sparse data
//==============================================================================
//
// Based on:
//
// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2007 Shai Shalev-Shwartz.
//  All Rights Reserved.
//==============================================================================
// File Name: WeightVector.h
// Written by: Shai Shalev-Shwartz (28.01.07)
// efficiently implements a weight vector for sparse data
//==============================================================================

#ifndef _AugWeightVector_H
#define _AugWeightVector_H

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <sstream>
#include <math.h>
#include "simple_sparse_vec_hash.h"

#define MAKE_MY_A_ONE_THRES 1e-256
#define EPS 0.0

#include <sys/time.h>

// eventually move this routine to its own file. Currently it clashes
// with the version in WeightVector.cc


/*
inline double get_time()
{
   struct timeval tv;
   if (gettimeofday(&tv, NULL)==0)
      return tv.tv_sec+((double)(tv.tv_usec))/1e6;
   else
      return 0.0;
}
*/

typedef double dbl2 __attribute__ ((vector_size(2*sizeof(double))));

class AugWeightVector {
  
 private:
  AugWeightVector(const AugWeightVector&);// disallowed
  
 public:

  // Construct an augmented Weight vector with dimension d
 AugWeightVector(uint dd) : d(dd), my_a(1.0), my_snorm(0.0), my_v(NULL), my_beta(0.0) {
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;

    make_my_a_one_time = 0;
  }

  // Construct a Weight vector with dimension d from a file
  AugWeightVector(uint dd,std::ifstream &model_file) : 
  d(dd), my_a(1.0), my_snorm(0.0), my_v(NULL), my_beta(0.0) {
    my_v = new double[d];
    for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    // assume first entry in model file is beta
    model_file >> my_beta;
    if (model_file.good()) {
      unsigned int ind = 0;
      model_file >> ind;
      while (model_file.good()) {
	char c;
	model_file >> c;
	double val;
	model_file >> val;
	my_v[ind] = val;
	my_snorm += val*val;
	model_file >> ind;
      }
    }
    make_my_a_one_time = 0;
  }

  // destructor
  ~AugWeightVector() { delete[] my_v;  }

  // this *= s
  // scale the beta component too, in this version
  void scale(double s) { 
    my_snorm *= (s*s); 
    if (s != 0.0) {
      my_a *= s; 
    } else {
      my_a = 1.0;
      for (uint i=0; i<d; ++i) my_v[i] = 0.0;
      my_beta = 0.0;
    }
  }

  // this *= s, but in this version the beta component is not changed.
  void scaleW(double s) { 
    my_snorm *= (s*s); 
    if (s != 0.0) {
      my_beta /= s;
      my_a *= s; 
    } else {
      my_beta *= my_a;
      my_a = 1.0;
      for (uint i=0; i<d; ++i) my_v[i] = 0.0;
    }
  }

  // w += s*x, beta += s*del_beta
  void add(simple_sparse_vector& x, double s);

  // this += s*x
  void add(AugWeightVector& x, double s);

  //  void updateL1(double ustep, simple_sparse_vector& x, double s);
  //  void updateL1(double ustep);

  void print(std::ostream& os);

  double operator[](uint i) {
    if (i < d) return (my_v[i] * my_a);
    return 0.0;
  }

  double beta() {
    return my_beta * my_a;
  }

  uint dimension() {return d;}

  // ||this||^2
  double snorm() {
    return my_snorm;
  }

  // copy other
  void operator= (const AugWeightVector& other) {
    if (d != other.d) {
      std::cerr << "Assigning WeightVector of size "
		<< other.d << " to a WeightVector of size "
		<< d << " is not permitted" << std::endl;
      exit(EXIT_FAILURE);
    }
    my_a = other.my_a; my_snorm = other.my_snorm;
    for (uint i=0; i<d; ++i) my_v[i] = other.my_v[i];
    my_beta = other.my_beta;
  }

  // inner product (using SSE2)
  double operator*(const AugWeightVector& other) {
    if(d != other.d) {
      std::cerr << "Inner product of AugWeightVector of size "
	        << other.d << " to a AugWeightVector of size "
                << d << " is not permitted" << std::endl;
      exit(EXIT_FAILURE);
    }
    /*
    double dp;
    double sums[2] = {0,0};
    dbl2 _sums = *(dbl2*)sums;
    double a[2];
    a[0] = a[1] = my_a*other.my_a;
    dbl2 _a = *(dbl2*)a;
    uint i=d;
    double *x = my_v;
    double *y = other.my_v;
    while (i>=2) {
      //dp += my_a*my_v[i]*other.my_a*other.my_v[i];
      dbl2 _x = *(dbl2*)x;
      dbl2 _y = *(dbl2*)y;
      _sums = _sums + _a * _x * _y;
      i-=2;
      x+=2;
      y+=2;
    }
    *(dbl2*)sums = _sums;
    dp = sums[0] + sums[1];

    while(--i >= 0)
      dp += a[0]*(*x++)*(*y++);
    */
    double dp = 0;
    double ap = my_a * other.my_a;
    for(uint i=0; i<d; ++i)
      dp += ap*my_v[i]*other.my_v[i];
    return dp;
  }

  // make_my_a_one
  // use it for forcing my_a to be 1.0
  void make_my_a_one() {
    if(fabs(my_a) >= MAKE_MY_A_ONE_THRES)
	return;

    //    double start_time = get_time();
    for (uint i=0; i<d; ++i) {
      my_v[i] *= my_a;
    }
    my_beta *= my_a;
    my_a = 1.0;

    //    make_my_a_one_time += (get_time() - start_time);
  }

  uint nnz() {
    uint cnt=0;
    for (uint i=0; i<d; ++i) {
      if(fabs(my_v[i]) > EPS) ++cnt;
    }
    return cnt;
  }

 private:  
  // The internal representation of w is as w = a*v where:
  uint d;
  long double my_a; 
  long double my_snorm;
  double* my_v;
  double my_beta;

 public:
  double make_my_a_one_time;
};


//-----------------------------------------------------------------------------
/** Operator * for vector-vector multiplication
    @param u A reference to a simple_sparse_vector
    @param v A reference to a WeightVector
    @return The product (double)
*/
double operator* (simple_sparse_vector& u, AugWeightVector& v);

//-----------------------------------------------------------------------------
/** Operator * for vector-vector multiplication
    @param v A reference to a WeightVector
    @param u A reference to a simple_sparse_vector
    @return The product (double)
*/
double operator* (AugWeightVector& v, simple_sparse_vector& u);




#endif
