// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2007 Shai Shalev-Shwartz.
//  All Rights Reserved.
//
// Main for the PEGASOS algorithm.
//
//
// 
// Pegasos: Fly high in the primal 
// main file
//
// Modifed by SKL and SJW, 2009.

//*****************************************************************************
// Included Files
//*****************************************************************************
#include "cmd_line.h"
#include "simple_sparse_vec_hash.h"
#include "pegasos_optimize.h"


void print_results (int max_iter, double lambda, double trainTime, 
		    double calc_obj_time,  double norm_value,
		    double loss_value, double zero_one_error,
		    double obj_value, double test_loss,
		    double test_error);

double target_err = 0.0;

int main(int argc, char** argv) {


  // -------------------------------------------------------------
  // ---------------------- Parse Command Line -------------------
  // -------------------------------------------------------------

  std::string data_filename;
  std::string test_filename;
  std::string model_filename;
  std::string experiments_file = "noExperimentsFile";
  double lambda = 1.0;
  int max_iter = 10;
  int exam_per_iter = 1;
  double gamma = 1.0;
//  double ignore_ratio = 0.0;
  double sol_eps = 1e-12;
  int rank = 32;
  int algo = 0;
  int b = 0;
  int avg_no = 1;
  double theta = 1.0;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("Pegasos");
  cmdline.add_master_option("<data-file>", &data_filename);
  cmdline.add("-algo", "algorithm to use {0,1,2,3,4,5,6,7} (default: 0)\n\
0: 1/t rule, linear\n\
1: 1/sqrt(t) rule, linear\n\
2: 1/t rule, Gaussian\n\
3: 1/sqrt(t) rule, Gaussian\n\
4: 1/t rule, Gaussian, low-dim\n\
5: 1/sqrt(t) rule, Gaussian, low-dim\n\
6: 1/t rule, Gaussian, random FT\n\
7: 1/sqrt(t) rule, Gaussian, random FT\
", &algo, 0);
  cmdline.add("-b", "use itercept {0,1} (default: 0)", &b, 0);
  cmdline.add("-lambda", "regularization parameter (default = 0.001)", &lambda, 0.001);
  cmdline.add("-iter", "number of iterations (default = 100/lambda)", &max_iter, int(100/lambda));
  cmdline.add("-k", "size of block for stochastic gradient (default = 1)", &exam_per_iter, 1);
  cmdline.add("-g", "(nonliner): kernel bandwidth of the Gaussian kernel (if used, default = 1)", &gamma, 1.0);
  cmdline.add("-rank", "(nonliner): rank of the Nysrom approximation (default = 32)", &rank, 32);
//  cmdline.add("-ignore", "(nonlinear): the ratio of the absolute value of the solution to the bounds, to ignore for error computation (default 0.0)", &ignore_ratio, 0.0);
  cmdline.add("-sol_eps", "(nonlinear): the absolute value threshold of the solution to ignore for error computation (default 1e-12)", &sol_eps, 1e-12);
  cmdline.add("-avg_no", "(nonlinear): the number of last iterates to be averaged (Default: 1)", &avg_no, 1);
  cmdline.add("-modelFile","name of model file (default = noModelFile)", &model_filename,"noModelFile");
  cmdline.add("-testFile","name of test data file (default = noTestFile)", &test_filename,"noTestFile");
  cmdline.add("-target_err", "target test error", &target_err, 0.0);
  cmdline.add("-theta", "steplength scalar theta (default: 1.0)", &theta, 1.0);
//   cmdline.add("-experiments","name of experiments spec. file", 
// 	      &experiments_file,"noExperimentsFile");

  int rc = cmdline.parse(argc, argv);
  if (rc < 2) {
    cmdline.print_help();
    return EXIT_FAILURE;
  }

  // -------------------------------------------------------------
  // ---------------------- read the data ------------------------
  // -------------------------------------------------------------
  uint dimension = 0;
  std::vector<simple_sparse_vector> Dataset;
  std::vector<int> Labels;
  double readingTime, sparsity;
  ReadData(data_filename,Dataset,Labels,dimension,sparsity,readingTime);

  std::cout << " Num Features= " << dimension 
	    << "\n Num Examples = " << Labels.size() 
	    << "\n Sparsity = " << sparsity
	    << "\n lambda = " << lambda
	    << std::endl;

  uint testDimension = 0;
  std::vector<simple_sparse_vector> testDataset;
  std::vector<int> testLabels;
  double testReadingTime, testSparsity;
  if (test_filename != "noTestFile") {
    ReadData(test_filename,testDataset,testLabels,testDimension,testSparsity,testReadingTime);
    std::cout << " Num Features= " << testDimension 
	      << "\n Num Examples = " << testLabels.size() 
	      << "\n Sparsity = " << testSparsity
	      << std::endl;

  } else {
    testReadingTime = 0;
  }
  std::cerr << readingTime+testReadingTime << " = Time for reading the data" <<  std::endl;

  // choose a random seed
  srand(time(NULL));


  // -------------------------------------------------------------
  // ---------------------- Experiments mode ---------------------
  // -------------------------------------------------------------
  if (experiments_file != "noExperimentsFile") {
    run_experiments(experiments_file,Dataset,Labels,dimension,
		    testDataset,testLabels);
    return(EXIT_SUCCESS); 
  }

  // pick which ones to run
  int runLearnReturnLast=0, runLearn=0, runLearnAndValidate=0, 
    runLearn_invsqrtt=0,
    runLearnRobust_invsqrtt=0,
    //runLearnRobust_const=0,
    //runLearnRobustAug_invt=0,
    runLearnRobustAug_invsqrtt=0,
    //runLearnRobustAug_const=0,
    runLearnNystrom=0,
    runLearnNystromRobust=0,
    runLearnNystromLowdim=0,
    runLearnNystromRobustLowdim=0,
    runLearnRandomFT=0,
    runLearnRobustRandomFT=0;

  //
  // Determine which algorithm to run
  //
  switch(algo) {
    case 0:	// 1/t rule, linear kernel
	if(b==0) runLearnReturnLast = 1;
	else {
	    //runLearnReturnLastAug = 1;
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

    case 1:	// 1/sqrt(t) rule, linear kernel
	if(b==0) runLearnRobust_invsqrtt = 1;
	else	 runLearnRobustAug_invsqrtt = 1;
	break;

    case 2:	// 1/t rule, Gaussian kernel
	if(b==0) runLearnNystrom = 1;
	else {
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

    case 3:	// 1/sqrt(t) rule, Gaussian kernel
	if(b==0) runLearnNystromRobust = 1;
	else {
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

    case 4:
	if(b==0) runLearnNystromLowdim = 1;
	else {
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

    case 5:
	runLearnNystromRobustLowdim = 1;
	break;

    case 6:
	if(b==0) runLearnRandomFT = 1;
	else {
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

     case 7:
	if(b==0) runLearnRobustRandomFT = 1;
	else {
	    std::cout << "Algorithm is not implemented yet" << std::endl;
	    return(EXIT_SUCCESS);
	}
	break;

    default:
	std::cout << "Unrecognized algorithm number is selected: " << algo << std::endl;
	return(EXIT_SUCCESS);
  }

  // -------------------------------------------------------------
  // ---------------------- Main Learning function ---------------
  // -------------------------------------------------------------
  double trainTime,calc_obj_time;
  double obj_value,norm_value,loss_value,zero_one_error,test_loss,test_error;
  
  // Pegasos, with averaging over the last "num_iter_to_average"
  // weight vectors

  if (runLearn) {
    std::cout << std::endl << "Learn - 1/(lambda*t) step rule" << std::endl;
    std::cout << ": eta_constant = m/(2 lambda)" << std::endl;
      uint num_iter_to_avg = 1;
      int eta_rule_type = 0;
      double eta_constant = 1.0;
      Learn(Dataset,Labels,dimension,testDataset,testLabels,
	    lambda,max_iter,exam_per_iter,
	    num_iter_to_avg,model_filename,
	    trainTime,calc_obj_time,obj_value,norm_value,loss_value,zero_one_error,
	    test_loss,test_error,
	    eta_rule_type,eta_constant, 0, 0.0);
      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
    }

  // run with sqrt step length.

  if (runLearn_invsqrtt) {
    std::cout << std::endl << "Learn - (eta_constant)/(sqrt(t)) step rule" << std::endl;
    int eta_rule_type = 1;
    double eta_constant = 1.0;
    uint num_iter_to_avg = 100;
    Learn(Dataset,Labels,dimension,testDataset,testLabels,
	  lambda,max_iter,exam_per_iter,
	  num_iter_to_avg,model_filename,
	  trainTime,calc_obj_time,obj_value,norm_value,loss_value,zero_one_error,
	  test_loss,test_error,
	  eta_rule_type,eta_constant,0,0.0);
    print_results (max_iter, lambda, trainTime, 
		   calc_obj_time,  norm_value,
		   loss_value,  zero_one_error,
		   obj_value,  test_loss,
		   test_error);
  }

  if (runLearnAndValidate) {
      uint num_example_to_validate = max_iter/10;
      LearnAndValidate(Dataset,Labels,dimension,testDataset,testLabels,
		       lambda,max_iter,exam_per_iter,
		       num_example_to_validate,model_filename,
		       trainTime,calc_obj_time,obj_value,norm_value,
		       loss_value,zero_one_error,
		       test_loss,test_error,
		       0,0.0,0,0.0);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
    }
  
  // Standard Pegasos
  if (runLearnReturnLast) {
    std::cout << "LearnReturnLast - 1/(lambda*t) step rule" << std::endl;
    LearnReturnLast(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,exam_per_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      0,0.0,0,0.0);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
    }

  // Nonlinear SVM, with Nystrom sampling to construct low-rank factors
  if (runLearnNystrom) {
    std::cout << "LearnNystrom - 1/(lambda*t) step rule" << std::endl;
    LearnNystrom(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      0, 0.0, 0, 0.0,
		      rank, gamma, sol_eps, avg_no);
//		      rank, gamma, ignore_ratio);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
  }

  // Nonlinear SVM, with Nystrom sampling to construct low-rank factors
  if (runLearnNystromRobust) {
    std::cout << "LearnNystromRobust - 1/sqrt(t) step rule" << std::endl;
    int steplength_rule = 0; 
    // "0"= 1/sqrt(i) steplength rule.
    // "1"=constant steplength rule
    // "2"=theta/(lambda*i)" steplength rule, no averaging
    int projection_rule = 0;
    // "0"=project onto 2-ball of radius 1/sqrt(lambda)
    // "1"=project onto 2-ball of radius projection_constant
    // "2"=no projection.
    double projection_constant=0.0;
    // scaling factor for step

    LearnNystromRobust(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      steplength_rule, theta,
		      projection_rule, projection_constant,
		      rank, gamma, sol_eps, avg_no);
//		      rank, gamma, ignore_ratio);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
  }

  if (runLearnNystromLowdim) {
    std::cout << "LearnNystromLowdim - 1/(lambda*t) step rule" << std::endl;
    LearnNystromLowdim(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      0, 0.0, 0, 0.0,
		      rank, gamma, sol_eps, avg_no);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
  }

  if (runLearnNystromRobustLowdim) {
    std::cout << "LearnNystromRobustLowdim - 1/sqrt(t) step rule" << std::endl;
    int steplength_rule = 0; 
    // "0"= 1/sqrt(i) steplength rule.
    // "1"=constant steplength rule
    // "2"=theta/(lambda*i)" steplength rule, no averaging
    int projection_rule = 0;
    // "0"=project onto 2-ball of radius 1/sqrt(lambda)
    // "1"=project onto 2-ball of radius projection_constant
    // "2"=no projection.
    double projection_constant=0.0;
    // scaling factor for step

    LearnNystromRobustLowdim(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      steplength_rule, theta,
		      projection_rule, projection_constant,
		      rank, gamma, sol_eps, avg_no, b);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
  }

#if 1
  // Nonlinear SVM, with random sampling to construct low-rank kernel function approximations
  if (runLearnRandomFT) {
    std::cout << "LearnRandomFT - 1/(lambda*t) step rule" << std::endl;
    LearnRandomFT(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,exam_per_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      0 /*eta_rule_type*/,0.0,0,0.0, rank, gamma, avg_no);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
    }

  if (runLearnRobustRandomFT) {
    std::cout << "LearnRobustRandomFT - 1/sqrt(t) step rule" << std::endl;
    LearnRobustRandomFT(Dataset,Labels,dimension,testDataset,testLabels,
		      lambda,max_iter,exam_per_iter,
		      model_filename,
		      trainTime,calc_obj_time,obj_value,norm_value,
		      loss_value,zero_one_error,
		      test_loss,test_error,
		      0 /*eta_rule_type*/,0.0,0,0.0, rank, gamma, avg_no, b);

      print_results (max_iter, lambda, trainTime, 
		     calc_obj_time,  norm_value,
		     loss_value,  zero_one_error,
		     obj_value,  test_loss,
		     test_error);
    }
#endif

  return(EXIT_SUCCESS);
}


void print_results (int max_iter, double lambda, double trainTime, 
		    double calc_obj_time,  double norm_value,
		    double loss_value, double zero_one_error,
		    double obj_value, double test_loss,
		    double test_error)
{
/*
  std::cout << max_iter << " = Num of iterations\n"
	    << lambda << " = lambda\n"
	    << trainTime << " = Time for training\n" 
	    << calc_obj_time << " = Time to calculate objective\n" 
	    << norm_value  << " = |w| \n" 
	    << loss_value << " = avg Loss of solution\n"  
	    << zero_one_error  << " = zero-one error rate: training data\n" 
	    << obj_value << " = primal objective of solution\n" 
	    << test_loss << " = avg Loss over test\n"  
	    << test_error  << " = zero-one error rate: test data\n" 	    
	    <<  std::endl;
*/
  std::cout << std::endl
	    << "iter          = " << max_iter << std::endl
	    << "training time = " << trainTime  << std::endl
	    << "training err  = " << zero_one_error*100 << " %" << std::endl
	    << "test     err  = " << test_error*100  << " %" << std::endl
	    << "|w|    = " << norm_value  << std::endl
	    << "lambda = " << lambda << std::endl
	    << "obj    = " << obj_value << std::endl
//	    << "avg. loss:training = " << loss_value << std::endl
//	    << "avg. loss:testing  = " << test_loss << std::endl
	    << std::endl;
 
}




