// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2007 Shai Shalev-Shwartz.
//  All Rights Reserved.
//=============================================================================
// File Name: pegasos_optimize.cc
// implements the main optimization function of pegasos
//=============================================================================

#include "pegasos_optimize.h"

using namespace std;

extern double target_err;

#ifdef HUMAN
#define PRINTFREQ 1000	// 500
#else
#define PRINTFREQ num_examples/5
//#define PRINTFREQ max_iter/10;
#endif

// help function for getting runtime
/* Modified by Sangkyun Lee - see below
double get_runtime(void)
{
  clock_t start;
  start = clock();
  return(((double)start/(double)CLOCKS_PER_SEC));
}
*/

// Sangkyun Lee: More accurate version of get_runtime()
#include <sys/time.h>
double get_runtime()
{
   struct timeval tv;
   if (gettimeofday(&tv, NULL)==0)
      return tv.tv_sec+((double)(tv.tv_usec))/1e6;
   else
      return 0.0;
}

#include <time.h>
#define PI 3.141592653589793
// Random number generator
//#ifdef ALT_RAND
#include <stocc.h>
//#endif

#ifdef OMP
#include <omp.h>
//omp_set_num_threads(4);
#endif

// ------------------------------------------------------------//
// ---------------- OPTIMIZING --------------------------------//
// ------------------------------------------------------------//
void Learn(// Input variables
	   vector<simple_sparse_vector>& Dataset,
	   vector<int>& Labels,
	   uint dimension,
	   vector<simple_sparse_vector>& testDataset,
	   vector<int>& testLabels,
	   double lambda,int max_iter,int exam_per_iter,int num_iter_to_avg,
	   string& model_filename,
	   // Output variables
	   double& train_time,double& calc_obj_time,double& obj_value,
	   double& norm_value,double& loss_value,double& zero_one_error,
	   double& test_loss,double& test_error,
	   // additional parameters
	   int eta_rule_type , double eta_constant ,
	   int projection_rule, double projection_constant) {

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  cout << "\n in Learn with lambda = " << lambda 
	    << ", max_iter = " << max_iter << endl;

  // Initialization of classification vector
  WeightVector W(dimension);
  WeightVector AvgW(dimension);
  double avgScale = (num_iter_to_avg > max_iter) ? max_iter : num_iter_to_avg; 

  int printFreq = PRINTFREQ;

  train_time = 0;
  startTime = get_runtime();

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // learning rate
    double eta;
    if (eta_rule_type == 0) { // Pegasos eta rule
      eta = 1 / (lambda * (i+2)); 
    } else if (eta_rule_type == 1) { // Norma rule
      eta = eta_constant / sqrt(i+2);
    } else {
      eta = eta_constant;
    }

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;

    // calc sub-gradients
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r = ((int)rand()) % num_examples;

      // calculate prediction
      double prediction = W*Dataset[r];

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
      }
    }
 
    W.make_my_a_one();

    // scale w 
    W.scale(1.0 - eta*lambda);

    // and add sub-gradients
    for (uint j=0; j<grad_index.size(); ++j) {
      W.add(Dataset[grad_index[j]],grad_weights[j]);
    }

    // Project if needed
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection

    // and update AvgW
    if (max_iter <= num_iter_to_avg + i)
      AvgW.add(W, 1.0/avgScale);

    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(W.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
	double cur_loss = 1 - Labels[l]*(W * Dataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      loss_value /= num_examples;
      obj_value += loss_value;
      zero_one_error /= num_examples;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint l=0; l < testDataset.size(); ++l) {
    double cur_loss = 1 - testLabels[l]*(W * testDataset[l]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
 
      cout << " iteration " << i 
		<< " |w|=" << norm_value
                << " eta=" << eta
		<< " objective=" << obj_value 
		<< " zero-one error=" << zero_one_error 
		<< " test error=" << test_error 
		<< endl;


      startTime = get_runtime();
    }


  }

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
  norm_value = sqrt(AvgW.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double cur_loss = 1 - Labels[i]*(AvgW * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss/num_examples;
    obj_value += cur_loss/num_examples;
    if (cur_loss >= 1.0) zero_one_error += 1.0/num_examples;
  }

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double cur_loss = 1 - testLabels[i]*(AvgW * testDataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
    
    if(i % 100 == 0) { printf("."); fflush(NULL); }
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
  


  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    AvgW.print(model_file);
    model_file.close();
  }

}

void LearnAndValidate(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      int exam_per_iter,int num_example_to_validate,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant) {

  uint num_examples = Labels.size();

  double startTime = get_runtime();
  double endTime;
  

  // Initialization of classification vector
  WeightVector W(dimension);
  WeightVector BestW(dimension);
  double best_obj = 1.0; // the zero solution

  // create validation indices
  vector<uint> validate_indices(num_example_to_validate);
  for (uint i=0; i < validate_indices.size(); ++i)
    validate_indices[i] = ((int)rand()) % num_examples;

  // Choose s random indices 
  int s = 5; // corresponds to confidence of 0.9933
  int block_size = max_iter/s;
  vector<int> candidates(s);
  for (int i=0; i<s; ++i) {
    candidates[i] = block_size*i + ((int)rand()) % block_size;
  }
  candidates[s-1] = max_iter-1; // make sure we need all iterations
  int cur_block = 0;

  
  
  /*
  // specify print frequency
  int printFreq = PRINTFREQ;

  int M_samples=0;
  train_time = 0;
  startTime = get_runtime();

  evaluateMTime = get_runtime();

  if (eta_rule_type != 2) {
    // Preliminary calculations of constants used in steplength
    // settings. Base DX on the radius of the ball that w is restricted
    // to in the pegasos paper. Use sampling of the dataset to get the
    // bound on the square of the subgradient.
    DX = 1.0/sqrt(lambda);
    M_samples = num_examples / 100;

    if (M_samples <= 10 ) M_samples = 10;
    if (M_samples >= 100) M_samples = 100;
    for (int j=0; j < M_samples; j++) {
      // choose random example
      uint r = ((int)rand()) % num_examples;
      M += Dataset[r].snorm();
    }
    //M = sqrt(lambda + M/M_samples);
    M = sqrt(M/M_samples);
    //M = M*M;
    cout << " LearnAndValidate: M = " << M << " DX = " << DX << endl;
  }

  evaluateMTime = get_runtime() - evaluateMTime;
  cout << evaluateMTime << " = Time to evaluate M" << endl;

  double sum_eta=0.0, sum_eta_new=0.0;
  double eta, etamax, etamin;
  */

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {


     // learning rate
     double eta;
     if (eta_rule_type == 0) { // Pegasos eta rule
       eta = 1 / (lambda * (i+2));
     } else if (eta_rule_type == 1) { // Norma rule
       eta = eta_constant / sqrt(i+2); 
       // solve numerical problems
       /*
       if (projection_rule != 2)
         W.make_my_a_one();
       */
     } else {
       eta = eta_constant;
     }

    /*
    // step length
    double eta, etamax;

    if (eta_rule_type == 0) { 
      // "1/sqrt(i)" rule from Nemirovski Sec 2.
      eta = theta * DX / (M * sqrt(1.0*(i+1))); 
      etamax = theta * DX / (M * sqrt(1.0*(max_iter/4))); 
      if (eta > etamax) 
	eta = etamax;
      //eta = 1.0 / sqrt(lambda*(i+1)); 
    } else if (eta_rule_type == 1) { 
      // "constant steplength" rule from Nemirovski Sec 2.2
      eta = theta * DX / (M * sqrt(1.0*max_iter));
      // solve numerical problems
      //if (projection_rule != 2)
      //  W.make_my_a_one();
    } else {
      eta = eta_constant;
    }
    */

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;

    // calc sub-gradients
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r = ((int)rand()) % num_examples;

      // calculate prediction
      double prediction = W*Dataset[r];

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
      }
    }
 
    W.make_my_a_one();

    // scale w 
    W.scale(1.0 - eta*lambda);

    // and add sub-gradients
    for (uint j=0; j<grad_index.size(); ++j) {
      W.add(Dataset[grad_index[j]],grad_weights[j]);
    }

    // Project if needed
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection


    // and validate
    if (i == candidates[cur_block]) {
      double obj = 0.0;
      for (uint j=0; j < validate_indices.size(); ++j) {
	uint ind = validate_indices[j];
	double cur_loss = 1 - Labels[ind]*(W * Dataset[ind]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	obj += cur_loss;
      }
      obj /= validate_indices.size();
      obj += lambda/2.0*W.snorm();
      if (obj <= best_obj) {
	//	cerr << "obj of " << i << " (candidates[" 
	//		  << cur_block << "]) = " << obj << endl;
	BestW = W;
	best_obj = obj;
      }
      cur_block++;
    }
  }


  // update timeline
  endTime = get_runtime();
  train_time = endTime - startTime;
  startTime = get_runtime();

  // Calculate objective value
  norm_value = sqrt(BestW.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double cur_loss = 1 - Labels[i]*(BestW * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss/num_examples;
    obj_value += cur_loss/num_examples;
    if (cur_loss >= 1.0) zero_one_error += 1.0/num_examples;
  }

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double cur_loss = 1 - testLabels[i]*(BestW * testDataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
  


  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    BestW.print(model_file);
    model_file.close();
  }

}


void LearnReturnLast(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      int exam_per_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant) {

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  // Initialization of classification vector
  WeightVector W(dimension);

  cout << "\n in LearnReturnLast with lambda = " << lambda 
	    << ", max_iter = " << max_iter << endl;
  
  // specify print frequency
  int printFreq = PRINTFREQ;

  train_time = 0;
  startTime = get_runtime();

  uint nviol = 0;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // learning rate
    double eta;
    if (eta_rule_type == 0) { // Pegasos eta rule
      eta = 1 / (lambda * (i+2)); 
    } else if (eta_rule_type == 1) { // Norma rule
      eta = eta_constant / sqrt(i+2);
      // solve numerical problems
      //if (projection_rule != 2)
//	W.make_my_a_one();
    } else {
      eta = eta_constant;
    }

    // just set the scale factor to 1
    W.make_my_a_one();

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;

    // calc sub-gradients
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r = ((int)rand()) % num_examples;

      // calculate prediction
      double prediction = W*Dataset[r];

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	++nviol;
      }
    }
 
    // scale w 
    W.scale(1.0 - eta*lambda);

    // and add sub-gradients
    for (uint j=0; j<grad_index.size(); ++j) {
      W.add(Dataset[grad_index[j]],grad_weights[j]);
    }

    // Project if needed
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection

    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(W.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
	double cur_loss = 1 - Labels[l]*(W * Dataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      loss_value /= num_examples;
      obj_value += loss_value;
      zero_one_error /= num_examples;
      cout << " iteration " << i 
		<< " |w|=" << norm_value
		<< " objective=" << obj_value 
		<< " zero-one error=" << zero_one_error 
		<< endl;

      startTime = get_runtime();
    }


  }

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
  norm_value = sqrt(W.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double cur_loss = 1 - Labels[i]*(W * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double cur_loss = 1 - testLabels[i]*(W * testDataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
  


  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    W.print(model_file);
    model_file.close();
  }

  printf("nviol = %d (%.2f)\n", nviol, 100.0*nviol/max_iter);

}

inline double gaussian_kernel(simple_sparse_vector v1, simple_sparse_vector v2, double gamma)
{
    return ( exp(-gamma*( v1.snorm() - 2*(v1*v2) + v2.snorm() )));
}

extern "C" {
    /* SSYEVX prototype */
    extern void ssyevx_( char* jobz, char* range, char* uplo, int* n, float* a,
                       int* lda, float* vl, float* vu, int* il, int* iu, float* abstol,
                       int* m, float* w, float* z, int* ldz, float* work, int* lwork,
                       int* iwork, int* ifail, int* info );
    /* SSYEV prototype */
    extern void ssyev_( char* jobz, char* uplo, int* n, float* a, int* lda,
		       float* w, float* work, int* lwork, int* info);
    /* SGEMM prototype */
    extern int sgemm_(char *transa, char *transb, int *m, int * n, int *k, 
			float *alpha, float *a, int *lda, float *b, 
			int *ldb, float *beta, float *c, int *ldc);
    /* SGELS prototype */
    extern int sgels_(char *trans, int *m, int* n, int* nrhs, float* A, int* lda, 
			float* B, int* ldb, float* work, int* lwork, int* info);

    /* DSYEVX prototype */
    extern void dsyevx_( char* jobz, char* range, char* uplo, int* n, double* a,
	                int* lda, double* vl, double* vu, int* il, int* iu, double* abstol,
			int* m, double* w, double* z, int* ldz, double* work, int* lwork,
			int* iwork, int* ifail, int* info );
    /* SSYEV prototype */
    extern void dsyev_( char* jobz, char* uplo, int* n, double* a, int* lda,
		       double* w, double* work, int* lwork, int* info);

   /* DGEMM prototype */
    extern int dgemm_(char *transa, char *transb, int *m, int * n, int *k, 
			double *alpha, double *a, int *lda, double *b, 
			int *ldb, double *beta, double *c, int *ldc);
    /* DGELS prototype */
    extern int dgels_(char *trans, int *m, int* n, int* nrhs, double* A, int* lda, 
			double* B, int* ldb, double* work, int* lwork, int* info);
}

#ifdef KERPREC_DOUBLE
 #define KERDATA_TYPE double
 #define SYEVX dsyevx_
 #define SYEV dsyev_
 #define GEMM dgemm_
 #define GELS dgels_
#else
 #define KERDATA_TYPE float
 #define SYEVX ssyevx_
 #define SYEV ssyev_
 #define GEMM sgemm_
 #define GELS sgels_
#endif

typedef struct _sort_elem {
    int index;
    double val;
} sort_elem;

int sort_elem_compare(const void* _a, const void* _b)
{
    sort_elem* a = (sort_elem*)_a;
    sort_elem* b = (sort_elem*)_b;

    if(a->val < b->val) return -1;
    else if(a->val > b->val) return 1;
    else return 0;
}

int int_compare(const void* _a, const void* _b)
{
    return ( *(int*)_a - *(int*)_b );
}

#define SYEV_ABSTOL -1.0	// default tolerance

/*
MATRIX *invert_ltriangle_matrix(MATRIX *L)
// Given a lower triangular matrix L, computes inverse L^-1
{
  int i,j,k,n;
  double sum;
  MATRIX *I;
  
  if(L->m != L->n) {
    printf("ERROR: Matrix not quadratic. Cannot invert triangular matrix!\n");
    exit(1);
  }
  n=L->n;
  I=copy_matrix(L);

  for (i=0;i<n;i++) {
    I->element[i][i]=1.0/L->element[i][i];
    for (j=i+1;j<n;j++) {
      sum=0.0;
      for (k=i;k<j;k++) sum -= I->element[j][k]*I->element[k][i];
      I->element[j][i]=sum/L->element[j][j];
    }
  }

  return(I);
}

MATRIX *cholesky_matrix(MATRIX *A)
// Given a positive-definite symmetric matrix A[0..n-1][0..n-1], 
// this routine constructs its Cholesky decomposition, A = L · LT . 
// On input, only the upper triangle of A need be given; A is not modified. 
// The Cholesky factor L is returned in the lower triangle.
{
  int i,j,k,n;
  double sum;
  MATRIX *L;
  
  if(A->m != A->n) {
    printf("ERROR: Matrix not quadratic. Cannot compute Cholesky!\n");
    exit(1);
  }
  n=A->n;
  L=copy_matrix(A);

  for (i=0;i<n;i++) {
    for (j=i;j<n;j++) {
      for (sum=L->element[i][j],k=i-1;k>=0;k--) 
	sum -= L->element[i][k]*L->element[j][k];
      if (i == j) {
	if (sum <= 0.0) printf("Cholesky: Matrix not positive definite");
	L->element[i][i]=sqrt(sum);
      } 
      else L->element[j][i]=sum/L->element[i][i];
    }
  }
  // set upper triangle to zero
  for (i=0;i<n;i++) 
    for (j=i+1;j<n;j++) 
      L->element[i][j]=0;

  return(L);
}
*/

/*
MATRIX *cholesky_addcol_matrix(MATRIX *L,double *newcol)
// Given the Cholesky decomposition L of a positive-definite symmetric matrix A[0..n-1][0..n-1], 
// this routine constructs the Cholesky decomposition of A', which is A extended by the size n+1 column vector newcol. 
//   A'=[A   newcol]
//      [newcol^T  ]
// The Cholesky factor L is returned in the lower triangle. The L that is given to the function is free'd.
{
  int i,j,k,n;
  double sum;
  
  if(L->m != L->n) {
    printf("ERROR: Matrix not quadratic. Cannot compute Cholesky!\n");
    exit(1);
  }
  n=L->n;
  L=realloc_matrix(L,n+1,n+1);
  for (i=0;i<n+1;i++) 
    L->element[n][i]=newcol[i];

  for (k=0;k<n;k++) {
    L->element[n][k]/=L->element[k][k];
    for (j=k+1;j<n+1;j++) 
      L->element[n][j]-=L->element[n][k]*L->element[j][k];
  }
  L->element[n][n]=sqrt(L->element[n][n]);

  // set upper triangle to zero
  for (i=0;i<n;i++) 
    L->element[i][n]=0;

  return(L);
}
*/

void find_indep(KERDATA_TYPE *A, int n, double epsilon, double* indep)
// Given a positive-semidefinite symmetric matrix A[0..n-1][0..n-1], 
// this routine finds a subset of rows and colums that is linear independent. 
// To do this, it constructs the Cholesky decomposition, A = L · LT. 
// On input, only the upper triangle of A need be given; A IS MODIFIED. 
// The routine returns a vector in which non-zero elements indicate the linear independent subset. 
// epsilon is the amount by which the diagonal entry of L has to be greater than zero.
{
  register int i,j,k;
  double sum;
  KERDATA_TYPE *L = A;

  for (i=0;i<n;i++) {
    for (j=i;j<n;j++) {
      for (sum=L[i*n+j],k=i-1;k>=0;k--) 
	sum -= L[i*n+k]*L[j*n+k];
      if (i == j) {
	if (sum <= epsilon) sum=0.0;
	L[i*n+i]=sqrt(sum);
      } 
      else 
	if(L[i*n+i] == 0)
	  L[j*n+i]=0;
	else
	  L[j*n+i]=sum/L[i*n+i];
    }
  }
  // Gather non-zero diagonal elements
  for (i=0;i<n;i++) 
      indep[i]=L[i*n+i];
}

#define EIG_THRES 1e-16

int nystrom_factor(vector<simple_sparse_vector>& Dataset,
		    vector<int>& Labels,
		    int s /* no. samples */, int*& s_idx /* out */, int*& r_idx /* out */, KERDATA_TYPE*& V /* out */, KERDATA_TYPE*& Qsr /* out */, KERDATA_TYPE gamma, 
		    KERDATA_TYPE* pKrrSum=NULL /* out */)
{
    register int i,j;
    register KERDATA_TYPE dval;
    int m = Labels.size();

#ifdef ALT_RAND
  // Random number generator
  CRandomSFMT0 Random(time(NULL));
#endif

    //
    // Sample s indices from [1:m] (without replacement to avoid singularity)
    //
    //int* s_idx = (int*)malloc(s*sizeof(int));
    s_idx = (int*)malloc(s*sizeof(int));
    sort_elem* s_vals = (sort_elem*)malloc(m*sizeof(sort_elem));
    for (i=0; i<m; ++i) {
	s_vals[i].index = i;
#ifdef ALT_RAND
	s_vals[i].val = Random.Random();
#else
	s_vals[i].val = rand();
#endif
    }
    // Random permutution of s_vals
    qsort( (void*)s_vals, m, sizeof(sort_elem), sort_elem_compare );
    for (i=0; i<s; ++i) s_idx[i] = s_vals[i].index;
    // Sort s_idx in ascending order for performance
    qsort( (void*)s_idx, s, sizeof(int), int_compare );

#if 0
{
    // Find subset of samples which produce linearly independent kernel columns
    KERDATA_TYPE *Kss = (KERDATA_TYPE*)malloc(s*s*sizeof(KERDATA_TYPE));
    double* indep = (double*)malloc(s*sizeof(double));
    // Fill upper triangular part of KSS
    for (i=0; i<s; ++i) {
	for(j=i; j<s; ++j) {
	    Kss[i*s+j] = Labels[s_idx[i]]*Labels[s_idx[j]]*gaussian_kernel(Dataset[s_idx[i]], Dataset[s_idx[j]], gamma); 
	}
    }
    find_indep(Kss, s, 1e-6, indep);
    int new_s = 0;
    for (i=0; i<s; ++i) {
	if(indep[i] != 0.0) s_idx[new_s++] = s_idx[i];
    }
    s = new_s;
    free(indep);
    free(Kss);
}
#endif

    // Generate Cms, consisting of s columns of \hat K (column major)
    KERDATA_TYPE *Cms = (KERDATA_TYPE*)malloc(m*s*sizeof(KERDATA_TYPE));
    /*
    for (i=0; i<s; ++i) {
	int t=0;
	for(j=0; j<m; ++j) {
	    if (j==s_idx[t]) {
		++t;
		if (i>j) Cms[i*m+j] = Cms[j*m+i];
	    } else 
		Cms[i*m+j] = Labels[s_idx[i]]*Labels[j]*gaussian_kernel(Dataset[s_idx[i]], Dataset[j], gamma); 
	}
    }
    */
#ifdef OMP
#pragma omp parallel private(j)
{
#pragma omp for
#endif
    for (i=0; i<s; ++i) {
	for(j=0; j<m; ++j) {
	    Cms[i*m+j] = Labels[s_idx[i]]*Labels[j]*gaussian_kernel(Dataset[s_idx[i]], Dataset[j], gamma); 
	}
    }
#ifdef OMP
}
#endif
    // Generate symmetric Kss, filling entries of upper triangular part (colume major)
    KERDATA_TYPE* Kss = (KERDATA_TYPE*)malloc(s*s*sizeof(KERDATA_TYPE));
    //memset(Kss, 0x00, s*s*sizeof(KERDATA_TYPE));
    if(pKrrSum) *pKrrSum = 0.0;
#ifdef OMP
#pragma omp parallel private(j)
{
#pragma omp for
#endif
    for (i=0; i<s; ++i) {
	for(j=0; j<=i; ++j) {
	    dval = Cms[i*m+s_idx[j]];
	    Kss[i*s+j] = dval;
	    if(pKrrSum) *pKrrSum += fabs(dval);
	}
    }
#ifdef OMP
}
#endif
 
    if(pKrrSum) {
	dval = 0.0;
	for(i=0; i<s; ++i) dval += fabs(Kss[i*s+i]);
	*pKrrSum = (*pKrrSum)*2.0 - dval; //*pKrrSum = (*pKrrSum - dval)*2.0 + dval;
    }

    //
    // Compute eigenvectors/eigenvalues of Kss
    //
    int lda = s;
    int lwork = -1;
    int info;
    KERDATA_TYPE wkopt;
    KERDATA_TYPE *lambda = (KERDATA_TYPE*)malloc(s*sizeof(KERDATA_TYPE));	// eigenvalues found
    SYEV("V", "U", &s, Kss, &lda, lambda, &wkopt, &lwork, &info);
    lwork = (int)wkopt;
    KERDATA_TYPE* work = (KERDATA_TYPE*)malloc(lwork*sizeof(KERDATA_TYPE));
    SYEV("V", "U", &s, Kss, &lda, lambda, work, &lwork, &info);
    if (info > 0) {
	cout << "SYEV failed to compute eigen-decomposition" << endl;
	exit(-1);
    }
    printf("Eigenvalues: smallest/largest of %d = %6.4e/%6.4e\n", s, lambda[0], lambda[s-1]);
    free(work);
    
    // Compute number of (large enough) positive eigenvalues
    // Eigenvalues are stored in lambda in ascending order
    int r=0;
    for (i=s-1; i>=0; --i) {
	if(lambda[i] < EIG_THRES)
	    break;
	++r;
    }
    printf("Effective rank = %d\n", r);

    //
    // Compute an approximate factor V = Cms Z (diag(lambda))^-1
    //
    KERDATA_TYPE alpha = 1.0;
    KERDATA_TYPE beta = 0.0;

    // Construct Qsr = U_sr D_rr^{-1/2}
    //KERDATA_TYPE* Qsr = (KERDATA_TYPE*)malloc(s*r*sizeof(KERDATA_TYPE));
    Qsr = (KERDATA_TYPE*)malloc(s*r*sizeof(KERDATA_TYPE));
#ifdef OMP
#pragma omp parallel private(j)
{
#pragma omp for
#endif
    for(i=0; i<r; ++i) {
	int real_i = i+s-r;
	double inv_sqrt_lam = sqrt(1./lambda[real_i]);
	for(j=0; j<s; ++j) {
	    Qsr[i*s+j] = Kss[real_i*s+j] * inv_sqrt_lam;
	}
    }
#ifdef OMP
}
#endif

    free(lambda);
    free(Kss);

    // Compute Vcol = Cms*Qsr
    //KERDATA_TYPE* Vcol = (KERDATA_TYPE*)malloc(m*r*sizeof(KERDATA_TYPE));	// column-major temporary storage of V
    //GEMM("N", "N", &m, &r, &r, &alpha, Cmr, &m, Krr, &r, &beta, Vcol, &m);
#ifdef OMP
    V = (KERDATA_TYPE*)malloc(m*r*sizeof(KERDATA_TYPE));
#pragma omp parallel private(j)
{
#pragma omp for
    for(i=0; i<m; ++i) {
	for(j=0; j<r; ++j) {
	    V[i*r+j] = 0.;
	    for(register int k=0; k<s; ++k) {
		V[i*r+j] += Cms[k*m+i] * Qsr[j*s+k];
	    }
	}
    }
}
    free(Cms);
#else
    KERDATA_TYPE* Vcol = (KERDATA_TYPE*)malloc(m*r*sizeof(KERDATA_TYPE));	// column-major temporary storage of V
    GEMM("N", "N", &m, &r, &s, &alpha, Cms, &m, Qsr, &s, &beta, Vcol, &m);

//    free(Qsr);
    free(Cms);

    // Convert the storage of V as row-major, for performance
    V = (KERDATA_TYPE*)malloc(m*r*sizeof(KERDATA_TYPE));
    for(i=0; i<m; ++i) {
	for(j=0; j<r; ++j) {
	    V[i*r+j] = Vcol[j*m+i];
	}
    }
    free(Vcol);
#endif

    // This seems to be wrong.
    r_idx = (int*)malloc(r*sizeof(int));
    for (i=0; i<r; ++i)
	r_idx[i] = s_idx[s-r+i];
    free(s_vals);

#ifdef REPORT_NYSTROM_ERROR
    /*
    // for testing, compute VV^T
    KERDATA_TYPE* Kapprox = (KERDATA_TYPE*)malloc(m*m*sizeof(KERDATA_TYPE));
    dgemm_("N", "T", &m, &m, &r, &alpha, Vcol, &m, Vcol, &m, &beta, Kapprox, &m);
    */

    // compute Frobenious norm of Kfull - Kapprox
    KERDATA_TYPE fnorm = 0;
    KERDATA_TYPE ftmp;
    double vv;
    double diag_err=0.0;
    const int ksample = 1000;
    for(i=0; i<ksample; ++i) {
	for(j=0; j<=i; ++j) {
	    vv=0.;
	    for(register int k=0; k<r; ++k) vv += Vcol[k*m+i]*Vcol[k*m+j];
	    //ftmp = Kfull[i*m+j] - Kapprox[i*m+j];
	    //ftmp = Kapprox[i*m+j] - Labels[i]*Labels[j]*gaussian_kernel(Dataset[j], Dataset[i], gamma);
	    ftmp = vv - Labels[i]*Labels[j]*gaussian_kernel(Dataset[j], Dataset[i], gamma);
	    if(i==j)
		diag_err += ftmp*ftmp;
	    else
		fnorm += ftmp*ftmp;
	}
    }
    printf("Kernel approximation err (f-norm)= %6.4f\n", 2*fnorm + diag_err);
#endif

    return r;
}

// Solve gam = V'alp for alp, V is m by r
void recover_alp(KERDATA_TYPE* alp, KERDATA_TYPE* Vrow, WeightVector& gam, int* r_idx, int r, int m)
{
    KERDATA_TYPE* work = NULL;
    KERDATA_TYPE wkopt;
    int info;
    int nrhs = 1;
    int lwork;
    for(register int i=0; i<r; ++i)
	alp[i] = gam[i];
    /*
    KERDATA_TYPE* V = (KERDATA_TYPE*)malloc(m*r*sizeof(KERDATA_TYPE));
    for(register int i=0; i<r; ++i)
	for(register int j=0; j<m; ++j)
	    V[i*m+j] = Vrow[j*r+i];
    */
    KERDATA_TYPE* Vtrans = (KERDATA_TYPE*)malloc(r*r*sizeof(KERDATA_TYPE));
    for(register int j=0; j<r; ++j) {
	for(register int i=0; i<r; ++i) {
	    //V[i*r+j] = Vrow[r_idx[j]*r+i];
	    //Vtrans[j*r+i] = Vrow[r_idx[j]*r+i];
	    Vtrans[j*r+i] = Vrow[r_idx[j]*r+i];
	}
    }

    // Query workspace
    lwork = -1;
    //GELS("N", &r, &m, &nrhs, Vtrans, &r, alp, &m, &wkopt, &lwork, &info);
    //GELS("T", &r, &m, &nrhs, V, &r, alp, &m, &wkopt, &lwork, &info);
    GELS("N", &r, &r, &nrhs, Vtrans, &r, alp, &r, &wkopt, &lwork, &info);
    lwork = (int)wkopt;
    work = (KERDATA_TYPE*)malloc(lwork*sizeof(KERDATA_TYPE));
    //GELS("N", &r, &m, &nrhs, Vtrans, &r, alp, &m, work, &lwork, &info);
    //GELS("T", &r, &m, &nrhs, V, &r, alp, &m, work, &lwork, &info);
    GELS("N", &r, &r, &nrhs, Vtrans, &r, alp, &r, work, &lwork, &info);
    if (info > 0) {
	cout << "GELS failed" << endl;
	exit(-1);
    }

#if 1
    // Compute Va = Vtrans * alp
    KERDATA_TYPE alpha = 1.0;
    KERDATA_TYPE beta = 0.0;
    KERDATA_TYPE* Va = (KERDATA_TYPE*)malloc(r*sizeof(KERDATA_TYPE));
    //GEMM("N", "N", &r, &nrhs, &m, &alpha, Vtrans, &r, alp, &m, &beta, Va, &r);
    //GEMM("T", "N", &r, &nrhs, &m, &alpha, V, &m, alp, &m, &beta, Va, &r);
    GEMM("N", "N", &r, &nrhs, &r, &alpha, Vtrans, &r, alp, &r, &beta, Va, &r);
    // Compute error
    double err= 0.0;
    for(int i=0; i<r; ++i)
	err += (Va[i] - gam[i])*(Va[i] - gam[i]);
    printf("RMS errro = %f\n", sqrt(err)/r);
#endif

    //free(V);
    free(Vtrans);
    free(work);

    return;
}

//
// Scaled stochastic subgradient method for nonlinear SVMs (strongly convex case)
// 
// Settings: H = K, and use Nystrom sampling for constructing low-rank factors of K.
// Created by Sang Lee.
//

void LearnNystrom(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant,
		      int sample, double gamma, double sol_eps, int avg_no) {

  const int col_samples = sample;		// the number of samples to construct the approximation, >= rank.

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  double DX, Mmax, M = 0.0, Mbar;

  // Initialization of classification vector
  WeightVector alp(num_examples);
  WeightVector gam(sample);
  WeightVector* avg_alp;
  if (avg_no > 1)
      avg_alp = new WeightVector(num_examples);
  else
      avg_alp = &alp;

  double eta_sum = 0.0, new_sum;

#ifdef ALT_RAND
  // Random number generator
  CRandomSFMT0 Random(time(NULL));
#endif

  cout << "\n in LearnReturnLast_lowrank with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    << ", sample = " << sample
	    << ", col_samples = " << col_samples
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    << endl;

  // Construct low-rank factor V of the kernel matrix K s.t. K ~ VV'
  // by Nystrom sampling (with uniform distribution)
  KERDATA_TYPE *V, *Qsr;
  int *r_idx, *s_idx;
  KERDATA_TYPE pKssSum[1];

  startTime = get_runtime();
  //nystrom_factor(Dataset, Labels, col_samples, rank, V, gamma, pKssSum);
  int rank = nystrom_factor(Dataset, Labels, sample, s_idx, r_idx, V, Qsr, gamma, pKssSum);
  printf("Nystrom approximation time = %.4f\n", get_runtime() - startTime);

  // Estimate D_X
  DX = sqrt(pKssSum[0])/rank/lambda;
  Mmax = DX*DX;
 
  // specify print frequency
  //int printFreq = PRINTFREQ;

  train_time = 0;
  startTime = get_runtime();

  double Vgamma;
  double s, m_ratio; 	// scale
  uint grad_index=-1;
  double grad_weight;
  double eta;
  int d;

#ifdef PERMUTE
    sort_elem* s_vals = (sort_elem*)malloc(num_examples*sizeof(sort_elem));
#endif

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {
#ifdef PERMUTE
    if(i % num_examples == 0) {
	for (register int j=0; j<num_examples; ++j) {
	    s_vals[j].index = i;
	    s_vals[j].val = rand();
	}
	qsort( (void*)s_vals, num_examples, sizeof(sort_elem), sort_elem_compare );
    }
#endif

    // learning rate
    if (eta_rule_type == 0) { // Pegasos eta rule
      eta = 1 / (lambda * (i+2)); 
      //eta = 1. / (i+2); 
    } else if (eta_rule_type == 1) { // Norma rule
      eta = eta_constant / sqrt(i+2);
      // solve numerical problems
      //if (projection_rule != 2)
//	W.make_my_a_one();
    } else {
      eta = eta_constant;
    }

    // just set the scale factor to 1
    alp.make_my_a_one();

    // calc sub-gradients
    grad_weight = 0.;
    d = 0;
    {
      // choose random example
#ifdef PERMUTE
      uint r = s_vals[i].index;
#else
#ifdef ALT_RAND
      uint r = Random.IRandom(0, num_examples-1);
      //uint r = Random.IRandomX(0, num_examples);
#else
      uint r = ((int)rand()) % num_examples;
#endif
#endif

      // calculate prediction
      //double prediction = alp*Dataset[r];
      Vgamma = gam.inner(&V[r*rank],rank); // V_{r.} gam

      // calculate loss
      //double cur_loss = 1 - Labels[r]*Vgamma;
      double cur_loss = 1. - Vgamma;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	//grad_index.push_back(r);
	//grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	grad_index = r;
	grad_weight = eta;
	d = 1;
      }
    }

    s = 1.0 - eta*lambda;

    // Check projection condition
    Mbar = (s*s)*M;
    if (d)
	Mbar += ( 2.*eta*Vgamma + eta*eta );

    if (Mbar <= Mmax) {
	M = Mbar;

	alp.scale(s);
	gam.scale(s);

	// and add sub-gradients
	if (d) {
	    alp.add(grad_index, grad_weight);
	    gam.add(&V[grad_index*rank], grad_weight, rank);
	}
    }
    else {
	M = Mmax;
	m_ratio = sqrt(Mmax / Mbar);
	s *= m_ratio;

	alp.scale(s);
	gam.scale(s);

	// and add sub-gradients
	if (d) {
	    alp.add(grad_index, m_ratio * grad_weight);
	    gam.add(&V[grad_index*rank], m_ratio * grad_weight, rank);
	}
    }

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_alp->scale(eta_sum/new_sum);
	avg_alp->add(alp, eta/new_sum);
	eta_sum = new_sum;
    }
    
#if 0
    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(alp.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      //loss_value = -1.0;
      //zero_one_error = -1.0;
      test_error = -1.0;

      // sample training error
      const int tr_samples = 100;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (int r=0; r < tr_samples; ++r) {
	uint l = (uint)(rand() % Dataset.size());	
	double Kalp = 0.0;
	for(register uint k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(Dataset[l], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(Kalp);
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      //loss_value /= num_examples;
      loss_value /= tr_samples;
      obj_value += loss_value;
      //zero_one_error /= num_examples;
      zero_one_error /= tr_samples;

/*
  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	for(register int k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[i]*(Kalp);

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    nest_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
*/
      cout << " iteration " << i 
		<< " |alp|2=" << norm_value
		<< " |alp|1=" << alp.norm1()
		<< " objective (untrue)" << obj_value 
		<< " zero-one error (sample)=" << zero_one_error 
		<< " test error=" << test_error 
		<< endl;

      startTime = get_runtime();
    }
#endif
  }

  // update timeline
  endTime = get_runtime();

  train_time += (endTime - startTime);
  startTime = get_runtime();

  vector<uint> sv;

  // Report density of vector alpha
  int alp_nnz = 0;
  int alp_true_nnz = 0;
  for(uint i=0; i<num_examples; ++i) {
    //if(fabs(avg_alp[i]) > ignore_ratio*proj_const) ++alp_nnz;
    if(fabs((*avg_alp)[i]) > sol_eps) { 
	++alp_nnz;
	sv.push_back(i);
    }
    if(fabs((*avg_alp)[i]) > 0.0) ++alp_true_nnz;
  }
  cout << "Alp density = " << 1.0*alp_nnz/num_examples << ", true = " << 1.*alp_true_nnz/num_examples<< endl;

  // Calculate objective value
  norm_value = sqrt(alp.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;

  double train_err_time, test_err_time;
  train_err_time = get_runtime();
/*
  printf("Compute train error");
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double Kalp = 0.0;
    for(register uint k=0; k<num_examples; ++k) {
        if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(Dataset[i], Dataset[k], gamma);
    }
    //double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
    double cur_loss = 1 - Labels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  train_err_time = get_runtime() - train_err_time;
*/

  // Calculate test_loss and test_error
  printf("\nCompute test error");
  test_err_time = get_runtime();
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	//for(register uint k=0; k<num_examples; ++k) {
	for(uint k=0; k < sv.size(); ++k) {
	    //if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k]);
	    //if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	    //if(fabs(alp[k]) > sol_eps) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	    //if(fabs((*avg_alp)[k]) > sol_eps) Kalp += (*avg_alp)[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	    Kalp += (*avg_alp)[sv[k]]*Labels[sv[k]]*gaussian_kernel(testDataset[i], Dataset[sv[k]], gamma);
	}
	double cur_loss = 1 - testLabels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }

  printf("\n"); fflush(NULL);
 
  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    alp.print(model_file);
    model_file.close();
  }

  test_err_time = get_runtime() - test_err_time;

  printf("true train err = %f, time(%.2f)\n", zero_one_error, train_err_time);
  printf("true test err  = %f, time(%.2f)\n", test_error, test_err_time);

  /*
  {
      double precompute_time, train_time, test_time;
      double train_loss, train_err, test_loss, test_err;

      compute_error_random(alp, Dataset, Labels, dimension, testDataset, testLabels,
	                   precompute_time, train_time, test_time,
			   train_loss, train_err, test_loss, test_err,
			   30*rank, gamma, ignore_ratio, proj_const);
      printf("approx train err = %f\n", train_err);
      printf("approx test err  = %f\n", test_err);
      printf("approx time: precompute(%.2f), train(%.2f), test(%.2f)\n", precompute_time, train_time, test_time);
  }
  */

  free(V);
  free(r_idx);
}

//
// (Robust) Scaled stochastic subgradient method for nonlinear SVMs (weakly convex case)
// 
// Settings: H = K, and use Nystrom sampling for constructing low-rank factors of K.
// Created by Sang Lee.

void LearnNystromRobust(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int steplength_rule, double theta,
		      int projection_rule, double projection_constant,
		      int sample, double gamma, double sol_eps, int avg_no) {
//		      int rank, double gamma, double ignore_ratio) {

  //const int col_samples = sample;		// the number of samples to construct the approximation, >= rank.

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  double DX, DG, Mmax, M = 0.0, Mbar;

  // Initialization of classification vector
  WeightVector alp(num_examples);
  WeightVector gam(sample);
  WeightVector* avg_alp;
  if (avg_no  > 1)
      avg_alp = new WeightVector(num_examples);
  else
      avg_alp = &alp;

  double eta_sum = 0.0, new_sum;

#ifdef ALT_RAND
  // Random number generator
  CRandomSFMT0 Random(time(NULL));
#endif

  cout << "\n in LearnReturnLast_lowrank with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    << ", sample = " << sample
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    << endl;

  // Construct low-rank factor V of the kernel matrix K s.t. K ~ VV'
  // by Nystrom sampling (with uniform distribution)
  KERDATA_TYPE *V, *Qsr;
  KERDATA_TYPE pKssSum[1];

  startTime = get_runtime();
  int *r_idx, *s_idx;
  int rank = nystrom_factor(Dataset, Labels, sample, s_idx, r_idx, V, Qsr, gamma, pKssSum);
  printf("Nystrom approximation time = %.4f\n", get_runtime() - startTime);

  // Estimate D_X
  DX = sqrt(pKssSum[0])/rank/lambda;
  Mmax = DX*DX;
  
  // specify print frequency
  //int printFreq = PRINTFREQ;

  train_time = 0;
  startTime = get_runtime();

  double Vgamma;
  double s, m_ratio;	// scale
  double d;		// indicator {-1,0}
  uint grad_index =-1;
  double grad_weight;
  double eta;

  /*
  if (steplength_rule != 2) {
    // Preliminary calculations of constants used in steplength
    // settings. Base DX on the radius of the ball that w is restricted
    // to in the pegasos paper. Use sampling of the dataset to get the
    // bound on the square of the subgradient.
    M_samples = num_examples / 100;

    if (M_samples <= 10 ) M_samples = 10;
    if (M_samples >= 100) M_samples = 100;
    M = 0.0;
    // average gradient contributions from the summation part
    for (int j=0; j < M_samples; j++) {
      // choose random example
      uint r = ((int)rand()) % num_examples;
      M += sqrt(Dataset[r].snorm());
    }
    // scale, and ad dthe other ingredients to make M an upper bound
    // estimate for E(||g||)
    M /= M_samples; M = M+1;
    cout << " LearnRobust: M = " << M << " DX = " << DX << endl;
  } else {
    // M and DX are not needed for the 1/t steplength rule; just set
    // them to zero
    M = 0.0; DX = 0.0;
  }
  */

  // Estimate M
  // Cheat: for the Gaussian kernel, we have M \approx 1.0
  DG = 1.0;

  cout << "Dx = " << DX << ", DG = " << DG << endl;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    //double etamax;

    // step length
    if (steplength_rule == 0) { 
      // "1/sqrt(i)" rule from Nemirovski Sec 2.
      eta = theta * DX / (DG * sqrt(1.0*(i+2))); 
      /*
      // bound the steplength by... something
      etamax = DX / (M * sqrt(1.0*(max_iter/10))); 
      if (eta > etamax) 
	eta = etamax;
      */
    } else if (steplength_rule == 1) { 
      // "constant steplength" rule from Nemirovski Sec 2.2, scaled by
      // theta
      eta = theta * DX / (M * sqrt(1.0*max_iter));
      // solve numerical problems
      //if (projection_rule != 2)
    } else if (steplength_rule == 2) { 
      // theta/(lambda*i) steplength. Note that theta=1 appears ideal;
      // performance degrades rapidly away from this value, as
      // expected from the analysis in Nemirovski Sec 2.1. No need for
      // M or DX here.
      eta = theta / (lambda*(i+1));
    } else {
      cerr << "steplength_rule unrecognized - TERMINATE" << endl;
      exit(1);
    }

    // just set the scale factor to 1
    alp.make_my_a_one();

    // calc sub-gradients
    grad_weight = 0.;
    d = 0;
    {
#ifdef ALT_RAND
      uint r = Random.IRandom(0, num_examples-1);
      //uint r = Random.IRandomX(0, num_examples);
#else
      uint r = ((int)rand()) % num_examples;
#endif

      // calculate prediction
      //double prediction = alp*Dataset[r];
      Vgamma = gam.inner(&V[r*rank],rank); // V_{r.} gam

      // calculate loss
      double cur_loss = 1 - Vgamma;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	//grad_index.push_back(r);
	//grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	grad_index = r;
	grad_weight = eta;
	d = 1;
      }
    }
 
    s = 1.0 - eta*lambda;

    // Check projection condition
    Mbar = (s*s)*M;
    if (d==-1)
	Mbar += ( 2.*eta*Vgamma + eta*eta );

    if (Mbar <= Mmax) {
	M = Mbar;
	
	alp.scale(s);
	gam.scale(s);

	if (d) {
	    alp.add(grad_index, grad_weight);
	    gam.add(&V[grad_index*rank], grad_weight, rank);
	}
    }
    else {
	M = Mmax;
	m_ratio = sqrt(Mmax / Mbar);
	s *= m_ratio;

	alp.scale(s);
	gam.scale(s);

	if (d) {
	    alp.add(grad_index, m_ratio * grad_weight);
	    gam.add(&V[grad_index*rank], m_ratio * grad_weight, rank);
	}
    }

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_alp->scale(eta_sum/new_sum);
	avg_alp->add(alp, eta/new_sum);
	eta_sum = new_sum;
    }
    
#if 0
    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(alp.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      //loss_value = -1.0;
      //zero_one_error = -1.0;
      test_error = -1.0;

      // sample training error
      const int tr_samples = 100;
      loss_value = 0.0;
      zero_one_error = 0.0;
      //for (uint l=0; l < Dataset.size(); ++l) {
      for (int r=0; r < tr_samples; ++r) {
	uint l = (uint)(rand() % Dataset.size());	
	double Kalp = 0.0;
	for(register uint k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(Dataset[l], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(Kalp);
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      //loss_value /= num_examples;
      loss_value /= tr_samples;
      obj_value += loss_value;
      //zero_one_error /= num_examples;
      zero_one_error /= tr_samples;

/*
  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	for(register int k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[i]*(Kalp);

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
*/
      cout << " iteration " << i 
		<< " |alp|2=" << norm_value
		<< " |alp|1=" << alp.norm1()
		<< " objective (untrue)" << obj_value 
		<< " zero-one error (sample)=" << zero_one_error 
		<< " test error=" << test_error 
		<< endl;

      startTime = get_runtime();
    }
#endif
  }

  // update timeline
  endTime = get_runtime();

  train_time += (endTime - startTime);
  startTime = get_runtime();

  vector<uint> sv;

  // Report density of vector alpha
  int alp_nnz = 0;
  int alp_true_nnz = 0;
  for(uint i=0; i<num_examples; ++i) {
//    if(fabs(alp[i]) > ignore_ratio*proj_const) ++alp_nnz;
    if(fabs((*avg_alp)[i]) > sol_eps) {
	++alp_nnz;
	sv.push_back(i);
    }
    if(fabs((*avg_alp)[i]) > 0.0) ++alp_true_nnz;
  }
  cout << "Alp density = " << 1.0*alp_nnz/num_examples << ", true = " << 1.*alp_true_nnz/num_examples<< endl;

  // Calculate objective value
  norm_value = sqrt(alp.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;

  double train_err_time, test_err_time;
  train_err_time = get_runtime();

/*
  printf("Compute train error");
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double Kalp = 0.0;
    for(register uint k=0; k<num_examples; ++k) {
        if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(Dataset[i], Dataset[k], gamma);
    }
    //double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
    double cur_loss = 1 - Labels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  train_err_time = get_runtime() - train_err_time;
*/

  // Calculate test_loss and test_error
  printf("\nCompute test error");
  test_err_time = get_runtime();
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	//for(register uint k=0; k<num_examples; ++k) {
	for(register uint k=0; k<sv.size(); ++k) {
	    //if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k]);
	    //if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	    //if(fabs(alp[k]) > sol_eps) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	    Kalp += (*avg_alp)[sv[k]]*Labels[sv[k]]*gaussian_kernel(testDataset[i], Dataset[sv[k]], gamma);
	}
	double cur_loss = 1 - testLabels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
 
  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    alp.print(model_file);
    model_file.close();
  }

  test_err_time = get_runtime() - test_err_time;

  printf("\n"); fflush(NULL);

  printf("true train err = %f, time(%.2f)\n", zero_one_error, train_err_time);
  printf("true test err  = %f, time(%.2f)\n", test_error, test_err_time);

  /*
  {
      double precompute_time, train_time, test_time;
      double train_loss, train_err, test_loss, test_err;

      compute_error_random(alp, Dataset, Labels, dimension, testDataset, testLabels,
	                   precompute_time, train_time, test_time,
			   train_loss, train_err, test_loss, test_err,
			   30*rank, gamma, ignore_ratio, proj_const);
      printf("approx train err = %f\n", train_err);
      printf("approx test err  = %f\n", test_err);
      printf("approx time: precompute(%.2f), train(%.2f), test(%.2f)\n", precompute_time, train_time, test_time);
  }
  */

  free(V);
  free(r_idx);
}

//
// Stochastic subgradient method for nonlinear SVMs (strongly convex case)
// on a low-dimensional space.
// 
// Nystrom sampling for constructing low-rank factors.
// Created by Sang Lee.
//

void LearnNystromLowdim(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant,
		      int sample, double gamma, double sol_eps, int avg_no) {

  uint num_examples = Labels.size();

  double startTime;
  double endTime;
  double nystrom_time;

  double eta_sum = 0.0, new_sum;

#ifdef ALT_RAND
  // Random number generator
  CRandomSFMT0 Random(time(NULL));
#endif

  cout << "\n in LearnReturnLast_lowrank with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    << ", sample = " << sample
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    << endl;

  // Construct low-rank factor V of the kernel matrix K s.t. K ~ VV'
  // by Nystrom sampling (with uniform distribution)

  startTime = get_runtime();

  KERDATA_TYPE *V, *Qsr;
  int *r_idx, *s_idx;
  int rank = nystrom_factor(Dataset, Labels, sample, s_idx, r_idx, V, Qsr, gamma);
  nystrom_time = get_runtime() - startTime;
  printf("Nystrom approximation time = %.4f\n", nystrom_time);

  // Collect basis examples and erase the rest
  vector<simple_sparse_vector> Basis;
  for(int i=0; i<sample; ++i)
      Basis.push_back( Dataset[s_idx[i]] );
  Dataset.clear();
 
  // Initialization of classification vector
  WeightVector gam(rank);
  WeightVector* avg_gam;
  if (avg_no > 1)
      avg_gam = new WeightVector(rank);
  else
      avg_gam = &gam;

#ifdef ALL_AVG
  WeightVector gam_allavg(rank);
  double eta_sum_all=0.0;
#endif

//#ifdef TESTERR_PROGRESS
  printf("Computing test kernel\n");
  int test_size = testDataset.size();
  KERDATA_TYPE* KQ = NULL;

  if(test_size > 0) {

    KERDATA_TYPE* TK = (KERDATA_TYPE*)malloc(sample*sizeof(KERDATA_TYPE));
    KQ = (KERDATA_TYPE*)malloc(testDataset.size()*rank*sizeof(KERDATA_TYPE));

    for (uint i=0; i < testDataset.size(); ++i) {
#ifdef OMP
#pragma omp parallel
{
#pragma omp for
#endif
	for(int s=0; s<sample; ++s) {
	    //TK[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Dataset[s_idx[s]], gamma);
	    TK[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Basis[s], gamma);
	}
#ifdef OMP
}
#pragma omp parallel
{
#pragma omp for
#endif
	for(int r=0; r<rank; ++r) {
	    KQ[i*rank+r] = 0.0;
	    for(int s=0; s<sample; ++s) {
		KQ[i*rank+r] += TK[s]*Qsr[r*sample+s];
	    }
	}
#ifdef OMP
}
#endif
    }
    free(TK);
    // Remove test dataset
    testDataset.clear();
  }
  free(Qsr);


#ifdef TESTERR_PROGRESS
  bool do_avg = false;
#endif

  // specify print frequency
  int printFreq = PRINTFREQ;

  train_time = 0;
  startTime = get_runtime();

  double Vgamma;
  uint grad_index=-1;
  double grad_weight;
  double eta;
  int d;

#ifdef PERMUTE
    sort_elem* s_vals = (sort_elem*)malloc(num_examples*sizeof(sort_elem));
#endif

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {
#ifdef PERMUTE
    if(i % num_examples == 0) {
	for (register int j=0; j<num_examples; ++j) {
	    s_vals[j].index = i;
	    s_vals[j].val = rand();
	}
	qsort( (void*)s_vals, num_examples, sizeof(sort_elem), sort_elem_compare );
    }
#endif

    // learning rate
    if (eta_rule_type == 0) { // Pegasos eta rule
      eta = 1 / (lambda * (i+2)); 
      //eta = 1. / (i+2); 
    } else if (eta_rule_type == 1) { // Norma rule
      eta = eta_constant / sqrt(i+2);
      // solve numerical problems
      //if (projection_rule != 2)
//	W.make_my_a_one();
    } else {
      eta = eta_constant;
    }

    // just set the scale factor to 1
    gam.make_my_a_one();
#ifdef ALL_AVG
    gam_allavg.make_my_a_one();
#endif

    //
    // calc sub-gradients
    //
    grad_weight = 0.;
    // choose random example
#ifdef PERMUTE
    uint r = s_vals[i].index;
#else
#ifdef ALT_RAND
    uint r = Random.IRandom(0, num_examples-1);
    //uint r = Random.IRandomX(0, num_examples);
#else
    uint r = ((int)rand()) % num_examples;
#endif
#endif

    // calculate prediction
    //double prediction = alp*Dataset[r];
    Vgamma = gam.inner(&V[r*rank],rank); // V_{r.} gam

    // calculate loss
    //double cur_loss = 1 - Labels[r]*Vgamma;
    double cur_loss = 1. - Vgamma;
    if (cur_loss < 0.0) cur_loss = 0.0;

    // and add to the gradient
    d = 0;
    if (cur_loss > 0.0) {
	//grad_index.push_back(r);
	//grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	grad_index = r;
	grad_weight = eta;
	d = 1;
    }

    gam.scale(1.0 - eta*lambda);

    // and add sub-gradients
    if (d) {
        gam.add(&V[grad_index*rank], grad_weight, rank);
    }

    // Projection
    double norm2 = gam.snorm();
    if (norm2 > 1.0/lambda) {
	gam.scale(sqrt(1.0/(lambda*norm2)));
    }

#ifdef ALL_AVG
#if 0
    // calculate prediction
    Vgamma = gam_allavg.inner(&V[r*rank],rank); // V_{r.} gam

    // calculate loss
    cur_loss = 1. - Vgamma;
    if (cur_loss < 0.0) cur_loss = 0.0;

    // and add to the gradient
    d=0;
    if (cur_loss > 0.0) {
	grad_index = r;
	grad_weight = eta;
	d = 1;
    }

    eta_sum += eta;

    // to step and average
    //weighted average
    eta_scale = (eta/eta_sum)*eta;
    gam_allavg.scale(1.0 - lambda*eta_scale);
    if (d) {
        gam_allavg.add(&V[grad_index*rank], eta_scale, rank);
    }
    //simple average
    /*
    gam_allavg.scale(1.0 - lambda*eta*(i+1));
    if (d) {
        gam_allavg.add(&V[grad_index*rank], eta*(i+1), rank);
    }
    */

    // Projection
    norm2 = gam_allavg.snorm();
    if (norm2 > 1.0/lambda) {
	gam_allavg.scale(sqrt(1.0/(lambda*norm2)));
    }
#endif
    new_sum = eta_sum_all + eta;
    gam_allavg.scale(eta_sum_all/new_sum);
    gam_allavg.add(gam, eta/new_sum);
    eta_sum_all = new_sum;

    // Projection
    norm2 = gam_allavg.snorm();
    if (norm2 > 1.0/lambda) {
	gam_allavg.scale(sqrt(1.0/(lambda*norm2)));
    }

#endif

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_gam->scale(eta_sum/new_sum);
	avg_gam->add(gam, eta/new_sum);
	eta_sum = new_sum;
    }

#ifdef TESTERR_PROGRESS
    train_time += (get_runtime() - startTime);

    if (avg_no > 1 && !do_avg && (i+avg_no) % printFreq == 0) {
	do_avg = true;
	eta_sum = 0.0;
	avg_gam->reset();
    }

    if (do_avg) {
	new_sum = eta_sum + eta;
	avg_gam->scale(eta_sum/new_sum);
	avg_gam->add(gam, eta/new_sum);
	eta_sum = new_sum;
    }

    if (i%printFreq == 0) {

	cout << nystrom_time + train_time << " ";

#ifdef ALL_AVG
	test_error = 0.0;
	for (uint i=0; i < test_size; ++i) {
	    double Kgam = 0.0;
	    for(int r=0; r<rank; ++r) {
		Kgam += KQ[i*rank+r]*gam_allavg[r];
	    }
	    if (testLabels[i]*(Kgam) <= 0.0) test_error += 1.0;
	}

	if (test_size != 0) {
	    test_error /= test_size;
	}

	cout << test_error;
#endif

	test_error = 0.0;
	for (int i=0; i < test_size; ++i) {
	    double Kgam = 0.0;
	    for(int r=0; r<rank; ++r) {
		Kgam += KQ[i*rank+r]*(*avg_gam)[r];
		//Kgam += KQ[i*rank+r]*gam[r];
	    }
	    if (testLabels[i]*(Kgam) <= 0.0) test_error += 1.0;
	}

	if (test_size != 0) {
	    test_error /= test_size;
	}

	cout << test_error << endl;
	fflush(stdout);

	if(test_error <= target_err) {
	    startTime = get_runtime();
	    break;
	}

	do_avg = false;
    }

    startTime = get_runtime();
#endif

#if 0
    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(alp.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      //loss_value = -1.0;
      //zero_one_error = -1.0;
      test_error = -1.0;

      // sample training error
      const int tr_samples = 100;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (int r=0; r < tr_samples; ++r) {
	uint l = (uint)(rand() % Dataset.size());	
	double Kalp = 0.0;
	for(register uint k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(Dataset[l], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(Kalp);
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      //loss_value /= num_examples;
      loss_value /= tr_samples;
      obj_value += loss_value;
      //zero_one_error /= num_examples;
      zero_one_error /= tr_samples;

/*
  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	for(register int k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[i]*(Kalp);

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    nest_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
*/
      cout << " iteration " << i 
		<< " |alp|2=" << norm_value
		<< " |alp|1=" << alp.norm1()
		<< " objective (untrue)" << obj_value 
		<< " zero-one error (sample)=" << zero_one_error 
		<< " test error=" << test_error 
		<< endl;

      startTime = get_runtime();
    }
#endif
  }

  // update timeline
  endTime = get_runtime();

  train_time += (endTime - startTime);
  startTime = get_runtime();

  //vector<uint> sv;

/*
  // Recover alpha by solving gam = V'alp for alp restricted to r_idx
  KERDATA_TYPE* alp = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));
  recover_alp(alp, V, *avg_gam, r_idx, rank, num_examples);

  KERDATA_TYPE* sv_alp = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));
  KERDATA_TYPE* sv_idx = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));

  // Report density of vector alpha
  int nsv = 0;
  for(int i=0; i<rank; ++i) {
    printf("%f\n", alp[i]);

    if(fabs(alp[i]) > sol_eps) { 
	//sv.push_back(r_idx[i]);
	sv_alp[nsv] = alp[i];
	sv_idx[nsv] = r_idx[i];
	++nsv;
    }
  }
  cout << "Alp density = " << 1.0*nsv/num_examples << endl;
*/

  // Calculate objective value
  //norm_value = sqrt(alp.snorm());
  //obj_value = norm_value * norm_value * lambda / 2.0;
  norm_value = -1;
  obj_value = -1;

  double train_err_time, test_err_time;
  train_err_time = get_runtime();
/*
  printf("Compute train error");
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double Kalp = 0.0;
    for(register uint k=0; k<num_examples; ++k) {
        if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(Dataset[i], Dataset[k], gamma);
    }
    //double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
    double cur_loss = 1 - Labels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  train_err_time = get_runtime() - train_err_time;
*/

  // Calculate test_loss and test_error
  printf("\nCompute test error");
  test_err_time = get_runtime();
  test_loss = 0.0;
  test_error = 0.0;
  /*
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	//for(uint k=0; k < sv.size(); ++k) {
	for(uint k=0; k < nsv; ++k) {
	    Kalp += sv_alp[k]*Labels[sv_idx[k]]*gaussian_kernel(testDataset[i], Dataset[sv_idx[k]], gamma);
	}
	double cur_loss = 1 - testLabels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  */

  for (int i=0; i < test_size; ++i) {
     double Kgam = 0.0;
     for(int r=0; r<rank; ++r) {
	Kgam += KQ[i*rank+r]*(*avg_gam)[r];
	//Kgam += KQ[i*rank+r]*gam[r];
     }
    double cur_loss = 1 - testLabels[i]*(Kgam);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }

/*
  KERDATA_TYPE* KT = (KERDATA_TYPE*)malloc(sample*sizeof(KERDATA_TYPE));

  for (uint i=0; i < testDataset.size(); ++i) {
	double Kgam = 0.0;
	for(int s=0; s<sample; ++s) {
	    KT[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Dataset[s_idx[s]], gamma);
	}
	for(int r=0; r<rank; ++r) {
	    double Ktmp = 0.0;
	    for(int s=0; s<sample; ++s) {
		Ktmp += KT[s]*Qsr[r*sample+s];
	    }
	    Kgam += Ktmp*(*avg_gam)[r];
	}
	double cur_loss = 1 - testLabels[i]*(Kgam);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  free(KT);
*/
  if(KQ) free(KQ);

  if (test_size != 0) {
    test_loss /= test_size;
    test_error /= test_size;
  }

  printf("\n"); fflush(NULL);
 
  // finally, print the model to the model_file
  /*
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    alp.print(model_file);
    model_file.close();
  }
  */

  test_err_time = get_runtime() - test_err_time;

  //printf("train err = %f, time(%.2f)\n", zero_one_error, train_err_time);
  printf("test err  = %f, time(%.2f)\n", test_error, test_err_time);

  /*
  {
      double precompute_time, train_time, test_time;
      double train_loss, train_err, test_loss, test_err;

      compute_error_random(alp, Dataset, Labels, dimension, testDataset, testLabels,
	                   precompute_time, train_time, test_time,
			   train_loss, train_err, test_loss, test_err,
			   30*rank, gamma, ignore_ratio, proj_const);
      printf("approx train err = %f\n", train_err);
      printf("approx test err  = %f\n", test_err);
      printf("approx time: precompute(%.2f), train(%.2f), test(%.2f)\n", precompute_time, train_time, test_time);
  }
  */

  free(V);
  free(r_idx);
//  free(alp);
//  free(sv_idx);
//  free(sv_alp);
  if (avg_no > 1)
      delete avg_gam;
}

//
// (Robust) Stochastic subgradient method for nonlinear SVMs (weakly convex case)
// 
// Nystrom sampling for constructing low-rank factors of K.
// Created by Sang Lee.
//

void LearnNystromRobustLowdim(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int steplength_rule, double theta,
		      int projection_rule, double projection_constant,
		      int sample, double gamma, double sol_eps, int avg_no, int use_b=0) {

  uint num_examples = Labels.size();

  double startTime;
  double endTime;
  double nystrom_time;

  double DX, DG;

  double eta_sum = 0.0, new_sum;

#ifdef ALT_RAND
  // Random number generator
  CRandomSFMT0 Random(time(NULL));
#endif

  cout << "\n in LearnReturnLast_lowrank with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    << ", sample = " << sample
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    << ", theta = " << theta
	    << endl;

  // Construct low-rank factor V of the kernel matrix K s.t. K ~ VV'
  // by Nystrom sampling (with uniform distribution)
  KERDATA_TYPE *V, *Qsr;

  startTime = get_runtime();
  int *r_idx, *s_idx;
  int rank = nystrom_factor(Dataset, Labels, sample, s_idx, r_idx, V, Qsr, gamma);
  nystrom_time = get_runtime() - startTime;
  printf("Nystrom approximation time = %.4f\n", nystrom_time);

  // Collect basis examples and erase the rest
  vector<simple_sparse_vector> Basis;
  for(int i=0; i<sample; ++i)
      Basis.push_back( Dataset[s_idx[i]] );
  Dataset.clear();

  // Initialization of classification vector
  WeightVector gam(rank);
  WeightVector* avg_gam;
  if (avg_no  > 1)
      avg_gam = new WeightVector(rank);
  else
      avg_gam = &gam;

//#ifdef TESTERR_PROGRESS
  int test_size = testDataset.size();
  KERDATA_TYPE* KQ = NULL;

  if(test_size > 0) {
    printf("Computing test kernel\n");

    KERDATA_TYPE* TK = (KERDATA_TYPE*)malloc(sample*sizeof(KERDATA_TYPE));
    KQ = (KERDATA_TYPE*)malloc(testDataset.size()*rank*sizeof(KERDATA_TYPE));

    for (uint i=0; i < testDataset.size(); ++i) {
#ifdef OMP
#pragma omp parallel
{
#pragma omp for
#endif
	for(int s=0; s<sample; ++s) {
	    //TK[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Dataset[s_idx[s]], gamma);
	    TK[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Basis[s], gamma);
	}
#ifdef OMP
}
#pragma omp parallel
{
#pragma omp for
#endif
	for(int r=0; r<rank; ++r) {
	    KQ[i*rank+r] = 0.0;
	    for(int s=0; s<sample; ++s) {
		KQ[i*rank+r] += TK[s]*Qsr[r*sample+s];
	    }
	}
#ifdef OMP
}
#endif
    }
    free(TK);
    // Remove test dataset
    testDataset.clear();
  }
  free(Qsr);

#ifdef TESTERR_PROGRESS
  bool do_avg = false;
#endif

  // specify print frequency
  int printFreq = PRINTFREQ;

  // Estimate D_X
  if(!use_b)
    DX = 1./sqrt(lambda);
  else
    DX = sqrt(2./lambda);	// use B = 1/lambda
  
  train_time = 0;
  startTime = get_runtime();

  double Vgamma;
  double d;		// indicator {-1,0}
  uint grad_index =-1;
  double grad_weight;
  double eta;
  double b=0.;

  // Estimate DG
  int DG_samples = num_examples / 100;
  if (DG_samples <= 10 ) DG_samples = 10;
  if (DG_samples >= 100) DG_samples = 100;
  DG = 0.0;
  // average gradient contributions from the summation part
  for (int j=0; j < DG_samples; j++) {
    // choose random example
#ifdef ALT_RAND
    uint r = Random.IRandom(0, num_examples-1);
#else
    uint r = ((int)rand()) % num_examples;
#endif
    double tmpsum = 0.0, tmp;
    for(int k=0; k<rank; ++k) {
	tmp = V[r*rank+k];
	tmpsum += tmp*tmp;
    }
    DG += sqrt(tmpsum);
  }
  // scale, and ad dthe other ingredients to make M an upper bound
  // estimate for E(||g||)
  DG /= DG_samples; 
  if(use_b) DG += 1.0;
  //DG += 1.0;

  cout << "Dx = " << DX << ", DG = " << DG << " use_b = " << use_b<< endl;

  //double sum_gnorm = 0.0;

  const double step_scale = theta * DX / DG;
  //double etamax = 0.;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    //double etamax;

    // step length
    if (steplength_rule == 0) { 
      eta = step_scale / sqrt(i+2.0); 
      // bound the steplength by... something
      /*
      etamax = step_scale / sqrt(1.0*(max_iter/10)); 
      if (eta > etamax) 
	eta = etamax;
      */
    /*
    } else if (steplength_rule == 1) { 
      // "constant steplength" rule from Nemirovski Sec 2.2, scaled by
      // theta
      eta = theta * DX / (DG * sqrt(1.0*max_iter));
      // solve numerical problems
      //if (projection_rule != 2)
    } else if (steplength_rule == 2) { 
      // theta/(lambda*i) steplength. Note that theta=1 appears ideal;
      // performance degrades rapidly away from this value, as
      // expected from the analysis in Nemirovski Sec 2.1. No need for
      // M or DX here.
      eta = theta / (lambda*(i+1));
    */
    } else {
      cerr << "steplength_rule unrecognized - TERMINATE" << endl;
      exit(1);
    }

    // just set the scale factor to 1
    gam.make_my_a_one();

    // calc sub-gradients
    grad_weight = 0.;
    d = 0;
    {
#ifdef ALT_RAND
      uint r = Random.IRandom(0, num_examples-1);
#else
      uint r = ((int)rand()) % num_examples;
#endif

      // calculate prediction
      Vgamma = gam.inner(&V[r*rank],rank); // V_{r.} gam

      // calculate loss
      double cur_loss;
      if(!use_b)
	  cur_loss = 1 - Vgamma;
      else
	  cur_loss = 1 - Vgamma - Labels[r]*b;

      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index = r;
	grad_weight = eta;
	d = 1;
      }
    }
 
    gam.scale(1.0 - eta*lambda);

    if (d) {
	gam.add(&V[grad_index*rank], grad_weight, rank);
	if(use_b) b = b + grad_weight*Labels[grad_index];
    }

    // Projection
    double norm2 = gam.snorm();
    if(!use_b && norm2 > 1.0/lambda) {
	gam.scale(sqrt(1.0/(lambda*norm2)));
    } else if(use_b && norm2 + b*b > 2.0/lambda) {
	double norm_scale = sqrt(1.0/(lambda*norm2));
	gam.scale(norm_scale);
	b *= norm_scale;
    }

    /*
    double gnorm = gam.snorm();
    if(d) {
	double vsnorm = 0;
	for (int k=0; k<rank; ++k)
	    vsnorm += V[grad_index*rank+k]*V[grad_index*rank+k];
	gnorm +=  -Vgamma + vsnorm;
    }
    gnorm = sqrt(gnorm);
    sum_gnorm += gnorm;
    DG = sum_gnorm / (i+1);
    */

    /*
    cum_gam_snorm += norm2;
    if(++cum_gam_cnt == 100) {
	DG = sqrt(cum_gam_snorm + use_b*cum_gam_cnt) / cum_gam_cnt;
	cum_gam_snorm = 0.0;
	cum_gam_cnt = 0;
	//cout << "DG = " << DG << endl;
    }
    */

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_gam->scale(eta_sum/new_sum);
	avg_gam->add(gam, eta/new_sum);
	eta_sum = new_sum;
    }


#ifdef TESTERR_PROGRESS
    train_time += (get_runtime() - startTime);

    if (avg_no > 1 && !do_avg && (i+avg_no) % printFreq == 0) {
	do_avg = true;
	eta_sum = 0.0;
	avg_gam->reset();
    }

    if (do_avg) {
	new_sum = eta_sum + eta;
	avg_gam->scale(eta_sum/new_sum);
	avg_gam->add(gam, eta/new_sum);
	eta_sum = new_sum;
    }

    if (i%printFreq == 0) {

	test_error = 0.0;
	for (int t=0; t < test_size; ++t) {
	    double Kgam = 0.0;
	    for(int r=0; r<rank; ++r) {
		Kgam += KQ[t*rank+r]*(*avg_gam)[r];
	    }
	    if(!use_b) {
		if(testLabels[t]*(Kgam) <= 0.0) test_error += 1.0;
	    } else {
		if(testLabels[t]*(Kgam+b) <= 0.0) test_error += 1.0;
	    }
	}

	if (test_size != 0) {
	    test_error /= test_size;
	}

	cout << nystrom_time + train_time << " " << test_error << endl;
	fflush(stdout);

	if(test_error <= target_err) {
	    startTime = get_runtime();
	    break;
	}

	do_avg = false;

    }

    startTime = get_runtime();
#endif

#if 0
    // Calculate and print objective value  and error rates periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(alp.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      //loss_value = -1.0;
      //zero_one_error = -1.0;
      test_error = -1.0;

      // sample training error
      const int tr_samples = 100;
      loss_value = 0.0;
      zero_one_error = 0.0;
      //for (uint l=0; l < Dataset.size(); ++l) {
      for (int r=0; r < tr_samples; ++r) {
	uint l = (uint)(rand() % Dataset.size());	
	double Kalp = 0.0;
	for(register uint k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(Dataset[l], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(Kalp);
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      //loss_value /= num_examples;
      loss_value /= tr_samples;
      obj_value += loss_value;
      //zero_one_error /= num_examples;
      zero_one_error /= tr_samples;

/*
  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	for(register int k=0; k<num_examples; ++k) {
	    if(alp[k] != 0.0) Kalp += alp[k]*gaussian_kernel(testDataset[i], Dataset[k], gamma);
	}
	//double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
	double cur_loss = 1 - Labels[i]*(Kalp);

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
*/
      cout << " iteration " << i 
		<< " |alp|2=" << norm_value
		<< " |alp|1=" << alp.norm1()
		<< " objective (untrue)" << obj_value 
		<< " zero-one error (sample)=" << zero_one_error 
		<< " test error=" << test_error 
		<< endl;

      startTime = get_runtime();
    }
#endif
  }

  // update timeline
  endTime = get_runtime();

  train_time += (endTime - startTime);
  startTime = get_runtime();

  vector<uint> sv;

/*
  // Recover alpha by solving gam = V'alp for alp restricted to r_idx
  KERDATA_TYPE* alp = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));
  recover_alp(alp, V, *avg_gam, r_idx, rank, num_examples);

  KERDATA_TYPE* sv_alp = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));
  KERDATA_TYPE* sv_idx = (KERDATA_TYPE*)malloc(rank*sizeof(KERDATA_TYPE));

  // Report density of vector alpha
  int nsv = 0;
  for(int i=0; i<rank; ++i) {
    printf("%f\n", alp[i]);

    if(fabs(alp[i]) > sol_eps) { 
	//sv.push_back(r_idx[i]);
	sv_alp[nsv] = alp[i];
	sv_idx[nsv] = r_idx[i];
	++nsv;
    }
  }
  cout << "Alp density = " << 1.0*nsv/num_examples << endl;
*/

  // Calculate objective value
  //norm_value = sqrt(alp.snorm());
  //obj_value = norm_value * norm_value * lambda / 2.0;
  norm_value = -1;
  obj_value = -1;

  double train_err_time, test_err_time;
  train_err_time = get_runtime();

/*
  printf("Compute train error");
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double Kalp = 0.0;
    for(register uint k=0; k<num_examples; ++k) {
        if(fabs(alp[k]) > ignore_ratio*proj_const) Kalp += alp[k]*gaussian_kernel(Dataset[i], Dataset[k], gamma);
    }
    //double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
    double cur_loss = 1 - Labels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  train_err_time = get_runtime() - train_err_time;
*/

  // Calculate test_loss and test_error
  printf("\nCompute test error");
  test_err_time = get_runtime();
  test_loss = 0.0;
  test_error = 0.0;
  /*
  for (uint i=0; i < testDataset.size(); ++i) {
	double Kalp = 0.0;
	//for(uint k=0; k < sv.size(); ++k) {
	for(uint k=0; k < nsv; ++k) {
	    Kalp += sv_alp[k]*Labels[sv_idx[k]]*gaussian_kernel(testDataset[i], Dataset[sv_idx[k]], gamma);
	}
	double cur_loss = 1 - testLabels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  */

  for(int i=0; i<test_size; ++i) {
      double Kgam = 0.0;
      for(int r=0; r<rank; ++r) {
	  Kgam += KQ[i*rank+r]*(*avg_gam)[r];
      }
      double cur_loss;
      if(!use_b) cur_loss = 1. - testLabels[i]*(Kgam);
      else	 cur_loss = 1. - testLabels[i]*(Kgam + b);

      if(i%100==0) { printf("."); fflush(NULL); }

      if(cur_loss <0.) cur_loss = 0.;
      test_loss += cur_loss;
      if(cur_loss >= 1.) test_error += 1.0;
  }


  /*
  KERDATA_TYPE* KT = (KERDATA_TYPE*)malloc(sample*sizeof(KERDATA_TYPE));

  for (uint i=0; i < testDataset.size(); ++i) {
	double Kgam = 0.0;
	for(int s=0; s<sample; ++s) {
	    KT[s] = Labels[s_idx[s]]*gaussian_kernel(testDataset[i], Dataset[s_idx[s]], gamma);
	}
	for(int r=0; r<rank; ++r) {
	    double Ktmp = 0.0;
	    for(int s=0; s<sample; ++s) {
		Ktmp += KT[s]*Qsr[r*sample+s];
	    }
	    Kgam += Ktmp*(*avg_gam)[r];
	}
	double cur_loss = 1 - testLabels[i]*(Kgam);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  free(KT);
  */
  if(KQ) free(KQ);
  //free(Qsr);

  if (test_size != 0) {
    test_loss /= test_size;
    test_error /= test_size;
  }

  printf("\n"); fflush(NULL);
 
  // finally, print the model to the model_file
  /*
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    alp.print(model_file);
    model_file.close();
  }
  */

  test_err_time = get_runtime() - test_err_time;

  //printf("train err = %f, time(%.2f)\n", zero_one_error, train_err_time);
  printf("test err  = %f, time(%.2f)\n", test_error, test_err_time);

  /*
  {
      double precompute_time, train_time, test_time;
      double train_loss, train_err, test_loss, test_err;

      compute_error_random(alp, Dataset, Labels, dimension, testDataset, testLabels,
	                   precompute_time, train_time, test_time,
			   train_loss, train_err, test_loss, test_err,
			   30*rank, gamma, ignore_ratio, proj_const);
      printf("approx train err = %f\n", train_err);
      printf("approx test err  = %f\n", test_err);
      printf("approx time: precompute(%.2f), train(%.2f), test(%.2f)\n", precompute_time, train_time, test_time);
  }
  */

  if(use_b)
      printf("Intercept b = %f\n", b);

  free(V);
  free(r_idx);
//  free(alp);
//  free(sv_idx);
//  free(sv_alp);
  if (avg_no > 1)
    delete avg_gam;
}


//
// Classical SA method for nonlinear SVMs, using randomization idea of [Rahimi & Recht, 07]
// 
// Created by Sang Lee.
// exam_per_iter == 1 should hold for the current implementation
//

void FTRandomTest(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      int rank, double gamma) {

  uint num_examples = Labels.size();

  // Initialization of classification vector
  WeightVector W(rank);
  WeightVector* pOmega[rank];
  double* pB = new double[rank];
  double* one_omega = new double[dimension];
  int seed = time(0);
  StochasticLib1 SL(seed);
  CRandomSFMT0 Random(seed);
  double omega_std = sqrt(2.0*gamma);
  double Z_scale = sqrt(2./rank);

  // Generate omega vectors by random sampling
  for(int i=0; i<rank; ++i) {
      //printf("%d: ", i);
      // Sample one omega vector
      for(uint j=0; j<dimension; ++j) {
	one_omega[j] = SL.Normal(0, omega_std);
	//printf("%f ", one_omega[j]);
      }
      //printf("\n");

      pOmega[i] = new WeightVector(dimension, one_omega);
      //pB[i] = 2*PI*(1.0*rand()/RAND_MAX);
      pB[i] = 2.*PI*Random.Random();
  }

  double **z = (double**)malloc(rank*num_examples*sizeof(double*));
  for(uint i=0; i<num_examples; ++i) {
      z[i] = (double*)malloc(rank*sizeof(double));
      for(int j=0; j<rank; ++j) {
	z[i][j] = Z_scale * cos( (*pOmega[j])*Dataset[i] + pB[j] );
      }
  }

  double err = 0.0;
  double ksum = 0.0;
  double tmp;
  double ztz;
  int jobs = num_examples*(num_examples-1)/2;
  int report_int = (int)(jobs/20.0);
  int done = 0;
  for(uint i=0; i<num_examples; ++i) {
      for(uint j=i+1; j<num_examples; ++j) {
	ztz = 0.0;
	for(int k=0; k < rank; ++k) {
	    ztz += z[i][k]*z[j][k];
	}
	tmp = gaussian_kernel(Dataset[i], Dataset[j], gamma);
	ksum += tmp*tmp;
	tmp -= ztz;
	err += tmp*tmp;

        if(++done%report_int ==0) { printf("."); fflush(stdout); }
      }
  }
  printf("\n");
  err *= 2.;
  for(uint i=0; i<num_examples; ++i) {
	ztz = 0.0;
	for(int k=0; k < rank; ++k) {
	    ztz += z[i][k]*z[i][k];
	}
	tmp = gaussian_kernel(Dataset[i], Dataset[i], gamma);
	ksum += tmp*tmp;
	tmp -= ztz;
	err += tmp*tmp;
  }

  printf("relative kernel approximation error = %f\n", err/ksum);
}

void LearnRandomFT(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      int exam_per_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant,
		      int rank, double gamma, int avg_no) {

#if 0
 FTRandomTest(Dataset, Labels, dimension, rank, gamma);
#endif

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  cout << "\n in LearnRandomFT with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    //<< ", sample = " << sample
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    //<< ", theta = " << theta
	    << endl;

/*
  if(max_iter > num_examples) {
      cout << "max_iter is adjusted to num_examples: " << num_examples << endl;
      max_iter = num_examples;
  }
*/
  int do_random = 1;
  if(max_iter == num_examples) {
      cout << "Epoch=1: example randomization goes off" << endl;
      do_random = 0;
  }
 
  // Initialization of classification vector
  WeightVector W(rank);
  WeightVector* avg_W;
  if (avg_no > 1)
      avg_W = new WeightVector(rank);
  else
      avg_W = &W;
  double eta_sum = 0.0, new_sum;

  WeightVector* pOmega[rank];
  double* pB = new double[rank];
  double* one_omega = new double[dimension];
  int seed = time(0);
  StochasticLib1 SL(seed);
  CRandomSFMT0 Random(seed);
  double omega_std = sqrt(2.0*gamma);
  double Z_scale = sqrt(2./rank);

  train_time = 0;
  startTime = get_runtime();

  // Generate omega vectors by random sampling
  for(int i=0; i<rank; ++i) {
      for(uint j=0; j<dimension; ++j) {
	one_omega[j] = SL.Normal(0, omega_std);
      }
      pOmega[i] = new WeightVector(dimension, one_omega);
      pB[i] = 2.*PI*Random.Random();
  }

/*
  // Construct new feature vectors
  double** Z = new double*[num_examples];
  for(int i=0; i<num_examples; ++i) {
    Z[i] = new double[rank];
    for(int k=0; k < rank; ++k)
	Z[i][k] = Z_scale * cos( (*pOmega[k])*Dataset[i] + pB[k] );
  }
*/

#ifdef TESTERR_PROGRESS
  bool do_avg = false;
  int test_size = testDataset.size();
#endif

  // specify print frequency
  int printFreq = PRINTFREQ;

  uint nviol = 0;

  double *Z = new double[rank];
  double* Ztest = new double[rank];

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // learning rate
    double eta;
    if (eta_rule_type == 0) { // Pegasos eta rule
      eta = 1 / (lambda * (i+2)); 
    } else if (eta_rule_type == 1) { // Norma rule
      eta = eta_constant / sqrt(i+2);
      // solve numerical problems
      //if (projection_rule != 2)
//	W.make_my_a_one();
    } else {
      eta = eta_constant;
    }

    // just set the scale factor to 1
    W.make_my_a_one();

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;

    // calc sub-gradients
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example

      uint r;
      if(do_random)
	  r = ((int)rand()) % num_examples;
      else
	  r = (uint)(i % num_examples);

      // Construct projection to manifold
      for(int k=0; k < rank; ++k)
        Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[r] + pB[k] );

      // calculate prediction
      //double prediction = W.inner(Z[r], rank);
      double prediction = W.inner(Z, rank);

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	++nviol;
      }
    }
 
    // scale w 
    W.scale(1.0 - eta*lambda);

    // and add sub-gradients
    for (uint j=0; j<grad_index.size(); ++j) {
      //W.add(Dataset[grad_index[j]],grad_weights[j]);
      //W.add(Z[grad_index[j]], grad_weights[j], rank);
      W.add(Z, grad_weights[j], rank);
    }

    // Project if needed
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_W->scale(eta_sum/new_sum);
	avg_W->add(W, eta/new_sum);
	eta_sum = new_sum;
    }

 #ifdef TESTERR_PROGRESS
    train_time += (get_runtime() - startTime);

    if (avg_no > 1 && !do_avg && (i+avg_no) % printFreq == 0) {
	do_avg = true;
	eta_sum = 0.0;
	avg_W->reset();
    }

    if (do_avg) {
	new_sum = eta_sum + eta;
	avg_W->scale(eta_sum/new_sum);
	avg_W->add(W, eta/new_sum);
	eta_sum = new_sum;
    }

    if (i%printFreq == 0) {

	cout << train_time << " ";

	double test_time = get_runtime();

	test_error = 0.0;
	for (int i=0; i < test_size; ++i) {
	    for(int k=0; k < rank; ++k) {
	        Ztest[k] = Z_scale * cos( (*pOmega[k])*testDataset[i] + pB[k] );
	    }

	    if (testLabels[i]*avg_W->inner(Ztest,rank) <= 0.0) test_error += 1.0;
	}

	if (test_size != 0) {
	    test_error /= test_size;
	}

	cout << test_error << " " << get_runtime() - test_time << endl;
	fflush(stdout);

	if(test_error <= target_err) {
	    startTime = get_runtime();
	    break;
	}

	do_avg = false;
    }

    startTime = get_runtime();
#endif

    // Calculate and print objective value  and error rates periodically
/*
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(W.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
        // compute Z(Dataset[l])
        for(int k=0; k < rank; ++k) {
	  Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[l] + pB[k] );
	}
	//double cur_loss = 1 - Labels[l]*(W * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(W.inner(Z, rank));
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      loss_value /= num_examples;
      obj_value += loss_value;
      zero_one_error /= num_examples;
      cout << " iteration " << i 
		<< " |w|=" << norm_value
		<< " objective=" << obj_value 
		<< " zero-one error=" << zero_one_error 
		<< endl;

      startTime = get_runtime();
    }
*/

  } // end of main loop

  printf("End of the main loop\n");
  fflush(stdout);

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
/*
  norm_value = sqrt(avg_W->snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    // compute Z(Dataset[i])
    for(int k=0; k < rank; ++k) {
      Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[i] + pB[k] );
    }
    double cur_loss = 1 - Labels[i]*(avg_W->inner(Z, rank));
    //double cur_loss = 1 - Labels[i]*(W * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;
*/
  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;

//  double* Ztest = new double[rank];
  
  for (uint i=0; i < testDataset.size(); ++i) {
    // compute Z(testDataset[i])
    for(int k=0; k < rank; ++k) {
      Ztest[k] = Z_scale * cos( (*pOmega[k])*testDataset[i] + pB[k] );
    }
    //double cur_loss = 1 - testLabels[i]*(W * testDataset[i]); 
    double cur_loss = 1 - testLabels[i]*(avg_W->inner(Ztest, rank));
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }


  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    W.print(model_file);
    model_file.close();
  }

  printf("nviol = %d (%.2f)\n", nviol, 100.0*nviol/max_iter);

  delete[] one_omega;
  for(int i=0; i<rank; ++i)
      delete pOmega[i];
  delete[] pB;
  /*
  for(int i=0; i<num_examples; ++i)
      delete Z[i];
  */
  delete[] Z;
  delete[] Ztest;

}

void LearnRobustRandomFT(// Input variables
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      double lambda,int max_iter,
		      int exam_per_iter,
		      string& model_filename,
		      // Output variables
		      double& train_time,double& calc_obj_time,
		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      // additional parameters
		      int eta_rule_type , double eta_constant ,
		      int projection_rule, double projection_constant,
		      int rank, double gamma, int avg_no, int use_b=0) {

#if 0
 FTRandomTest(Dataset, Labels, dimension, rank, gamma);
#endif

  uint num_examples = Labels.size();

  double startTime;
  double endTime;

  double DX, DG;

  cout << "\n in LearnRobustRandomFT with lambda = " << lambda 
	    << ", max_iter = " << max_iter 
	    //<< ", sample = " << sample
	    << ", gaussian_gamma = " << gamma
	    << ", avg_no = " << avg_no
	    //<< ", theta = " << theta
	    << endl;

/*
  if(max_iter > num_examples) {
      cout << "max_iter is adjusted to num_examples: " << num_examples << endl;
      max_iter = num_examples;
  }
*/
  int do_random = 1;
  if(max_iter == num_examples) {
      cout << "Epoch=1: example randomization goes off" << endl;
      do_random = 0;
  }
 
  // Initialization of classification vector
  WeightVector W(rank);
  WeightVector* avg_W;
  if (avg_no > 1)
      avg_W = new WeightVector(rank);
  else
      avg_W = &W;
  double eta_sum = 0.0, new_sum;

  WeightVector* pOmega[rank];
  double* pB = new double[rank];
  double* one_omega = new double[dimension];
  int seed = time(0);
  StochasticLib1 SL(seed);
  CRandomSFMT0 Random(seed);
  double omega_std = sqrt(2.0*gamma);
  double Z_scale = sqrt(2./rank);

  // Estimate D_X
  if(!use_b)
    DX = 1./sqrt(lambda);
  else
    DX = sqrt(2./lambda);

  train_time = 0;
  startTime = get_runtime();

  // Generate omega vectors by random sampling
  for(int i=0; i<rank; ++i) {
      for(uint j=0; j<dimension; ++j) {
	one_omega[j] = SL.Normal(0, omega_std);
      }
      pOmega[i] = new WeightVector(dimension, one_omega);
      pB[i] = 2.*PI*Random.Random();
  }

/*
  // Construct new feature vectors
  double** Z = new double*[num_examples];
  for(int i=0; i<num_examples; ++i) {
    Z[i] = new double[rank];
    for(int k=0; k < rank; ++k)
	Z[i][k] = Z_scale * cos( (*pOmega[k])*Dataset[i] + pB[k] );
  }
*/

#ifdef TESTERR_PROGRESS
  bool do_avg = false;
  int test_size = testDataset.size();
#endif

  // specify print frequency
  int printFreq = PRINTFREQ;

  uint nviol = 0;

  double *Z = new double[rank];
  double* Ztest = new double[rank];

  // Estimate DG
  int DG_samples = num_examples / 100;
  if (DG_samples <= 10 ) DG_samples = 10;
  if (DG_samples >= 100) DG_samples = 100;
  DG = 0.0;
  // average gradient contributions from the summation part
  for (int j=0; j < DG_samples; j++) {
    // choose random example
#ifdef ALT_RAND
    uint r = Random.IRandom(0, num_examples-1);
#else
    uint r = ((int)rand()) % num_examples;
#endif
    double tmp=0.0;
    for(int k=0; k < rank; ++k){
        Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[r] + pB[k] );
	tmp += Z[k]*Z[k];
    }
    DG += sqrt(tmp);
  }
  // scale, and ad dthe other ingredients to make M an upper bound
  // estimate for E(||g||)
  DG /= DG_samples; 
  if(use_b) DG += 1.0;
  //DG += 1.0;
  cout << "Dx = " << DX << ", DG = " << DG << " use_b = " << use_b<< endl;

  const double step_scale = DX / DG;
  double eta;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // learning rate
    eta = step_scale / sqrt(i+2.);

    // just set the scale factor to 1
    W.make_my_a_one();

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;

    // calc sub-gradients
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r;
      if(do_random)
	  r = ((int)rand()) % num_examples;
      else
	  r = (uint)(i % num_examples);

      // Construct projection to manifold
      for(int k=0; k < rank; ++k)
        Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[r] + pB[k] );

      // calculate prediction
      //double prediction = W.inner(Z[r], rank);
      double prediction = W.inner(Z, rank);

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	++nviol;
      }
    }
 
    // scale w 
    W.scale(1.0 - eta*lambda);

    // and add sub-gradients
    for (uint j=0; j<grad_index.size(); ++j) {
      //W.add(Dataset[grad_index[j]],grad_weights[j]);
      //W.add(Z[grad_index[j]], grad_weights[j], rank);
      W.add(Z, grad_weights[j], rank);
    }

    // Project if needed
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection

    // Averaging
    if (avg_no > 1 && max_iter - i <= avg_no) {
	new_sum = eta_sum + eta;
	avg_W->scale(eta_sum/new_sum);
	avg_W->add(W, eta/new_sum);
	eta_sum = new_sum;
    }

 #ifdef TESTERR_PROGRESS
    train_time += (get_runtime() - startTime);

    if (avg_no > 1 && !do_avg && (i+avg_no) % printFreq == 0) {
	do_avg = true;
	eta_sum = 0.0;
	avg_W->reset();
    }

    if (do_avg) {
	new_sum = eta_sum + eta;
	avg_W->scale(eta_sum/new_sum);
	avg_W->add(W, eta/new_sum);
	eta_sum = new_sum;
    }

    if (i%printFreq == 0) {

	cout << train_time << " ";

	double test_time = get_runtime();

	test_error = 0.0;
	for (int i=0; i < test_size; ++i) {
	    for(int k=0; k < rank; ++k) {
	        Ztest[k] = Z_scale * cos( (*pOmega[k])*testDataset[i] + pB[k] );
	    }

	    if (testLabels[i]*avg_W->inner(Ztest,rank) <= 0.0) test_error += 1.0;
	}

	if (test_size != 0) {
	    test_error /= test_size;
	}

	cout << test_error << " " << get_runtime() - test_time << endl;
	fflush(stdout);

	if(test_error <= target_err) {
	    startTime = get_runtime();
	    break;
	}

	do_avg = false;
    }

    startTime = get_runtime();
#endif

    // Calculate and print objective value  and error rates periodically
/*
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      norm_value = sqrt(W.snorm());
      obj_value = norm_value * norm_value * lambda / 2.0;
      loss_value = 0.0;
      zero_one_error = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
        // compute Z(Dataset[l])
        for(int k=0; k < rank; ++k) {
	  Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[l] + pB[k] );
	}
	//double cur_loss = 1 - Labels[l]*(W * Dataset[l]); 
	double cur_loss = 1 - Labels[l]*(W.inner(Z, rank));
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) zero_one_error += 1.0;
      }
      loss_value /= num_examples;
      obj_value += loss_value;
      zero_one_error /= num_examples;
      cout << " iteration " << i 
		<< " |w|=" << norm_value
		<< " objective=" << obj_value 
		<< " zero-one error=" << zero_one_error 
		<< endl;

      startTime = get_runtime();
    }
*/

  } // end of main loop

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
/*
  norm_value = sqrt(avg_W->snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    // compute Z(Dataset[i])
    for(int k=0; k < rank; ++k) {
      Z[k] = Z_scale * cos( (*pOmega[k])*Dataset[i] + pB[k] );
    }
    double cur_loss = 1 - Labels[i]*(avg_W->inner(Z, rank));
    //double cur_loss = 1 - Labels[i]*(W * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  zero_one_error /= num_examples;
*/
  endTime = get_runtime();
  calc_obj_time = endTime - startTime;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;

//  double* Ztest = new double[rank];
  
  for (uint i=0; i < testDataset.size(); ++i) {
    // compute Z(testDataset[i])
    for(int k=0; k < rank; ++k) {
      Ztest[k] = Z_scale * cos( (*pOmega[k])*testDataset[i] + pB[k] );
    }
    //double cur_loss = 1 - testLabels[i]*(W * testDataset[i]); 
    double cur_loss = 1 - testLabels[i]*(avg_W->inner(Ztest, rank));
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }


  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    W.print(model_file);
    model_file.close();
  }

  printf("nviol = %d (%.2f)\n", nviol, 100.0*nviol/max_iter);

  delete[] one_omega;
  for(int i=0; i<rank; ++i)
      delete pOmega[i];
  delete[] pB;
  /*
  for(int i=0; i<num_examples; ++i)
      delete Z[i];
  */
  delete[] Z;
  delete[] Ztest;

}

void compute_error_random(// Input variables
		      WeightVector& alp,
		      vector<simple_sparse_vector>& Dataset,
		      vector<int>& Labels,
		      uint dimension,
		      vector<simple_sparse_vector>& testDataset,
		      vector<int>& testLabels,
		      // Output variables
		      double& precompute_time, double& calc_train_time, double& calc_test_time,
//		      double& obj_value,double& norm_value,
		      double& loss_value,double& zero_one_error,
		      double& test_loss,double& test_error,
		      int rank, double gamma, double ignore_ratio,
		      double proj_const) {

  uint num_examples = Labels.size();

  precompute_time = get_runtime();

  // Initialization of classification vector
  WeightVector* pOmega[rank];
  double* pB = new double[rank];
  double* one_omega = new double[dimension];
  int seed = time(0);
  StochasticLib1 SL(seed);
  double omega_std = sqrt(2.0*gamma);
  double* Z = new double[rank];				// TODO: one Z per a example per iter
  //double Z_scale = sqrt(2./rank);
  double Z_scale = sqrt(2./rank * sqrt(2.)*pow(PI,0.5*rank));

  // Generate omega vectors by random sampling
  for(int i=0; i<rank; ++i) {
      //printf("%d: ", i);
      // Sample one omega vector
      for(uint j=0; j<dimension; ++j) {
	one_omega[j] = SL.Normal(0, omega_std);
	//printf("%f ", one_omega[j]);
      }
      //printf("\n");

      pOmega[i] = new WeightVector(dimension, one_omega);
      pB[i] = 2*PI*(1.0*rand()/RAND_MAX);
  }

  // Compute sum_{i} <omega_j, alp_i * train_i> , j=1,...,rank for nonzero alps
  int *idx = (int*)malloc( num_examples * sizeof(int) );
  int nnz = 0;
  for(uint i=0; i<num_examples; ++i) {
    if(fabs(alp[i]) > ignore_ratio*proj_const)	idx[nnz++] = i;
  }
  double *alp_wx = (double*)malloc(rank*sizeof(double));
  for(int j=0; j<rank; ++j) {
      alp_wx[j] = 0.;
      for(int i=0; i<nnz; ++i) {
	alp_wx[j] += alp[idx[i]] * Z_scale * cos( (*pOmega[j])*Dataset[idx[i]] + pB[j] );
      }
  }

  precompute_time = get_runtime() - precompute_time;
  calc_train_time = get_runtime();

  printf("Compute train error");
  loss_value = 0.0;
  zero_one_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double Kalp = 0.0;
    for(int k=0; k<rank; ++k) {
	Kalp += Z_scale * cos( (*pOmega[k])*Dataset[i] + pB[k] ) * alp_wx[k];
    }
    //double cur_loss = 1 - Labels[l]*(alp * Dataset[l]); 
    double cur_loss = 1 - Labels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) zero_one_error += 1.0;
  }
  loss_value /= num_examples;
  //obj_value += loss_value;
  zero_one_error /= num_examples;

  calc_train_time = get_runtime() - calc_train_time;
  calc_test_time = get_runtime();

  // Calculate test_loss and test_error
  printf("\nCompute test error");
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double Kalp = 0.0;
    for(int k=0; k<rank; ++k) {
        Kalp += Z_scale * cos( (*pOmega[k])*testDataset[i] + pB[k] ) * alp_wx[k];
    }
    double cur_loss = 1 - testLabels[i]*(Kalp);

    if(i % 100 == 0) { printf("."); fflush(NULL); }

    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  printf("\n");
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }

  calc_test_time = get_runtime() - calc_test_time;

  delete[] one_omega;
  for(int i=0; i<rank; ++i)
      delete pOmega[i];
  delete[] pB;
  delete[] Z;
  free(idx);
  free(alp_wx);

}

// Implementation of Stochastic Approximation approaches to the primal
// formulation of SVM. Follows Sections 2.1 and 2.2 of the paper
// "Robust Stochastic Approximation Approach to Stochastic
// Programming" by A. Nemirovski, A. Juditsky, G. Lan, and A. Shapiro,
// SIAM J. Optimization 19 (2009), pp. 1574-1609.

// coded by SJW 2/14/09 and after.
// More changes started 9/8/09.

void LearnRobust(// Inputs
		 vector<simple_sparse_vector>& Dataset,
		 vector<int>& Labels,
		 uint dimension,
		 vector<simple_sparse_vector>& testDataset,
		 vector<int>& testLabels,
		 double lambda, int max_iter,
		 int exam_per_iter, double frac_iter_to_avg,
		 string& model_filename,
		 // Outputs
		 double& train_time,double& calc_obj_time,
		 double& obj_value,double& norm_value,
		 double& loss_value,double& train_error,
		 double& test_loss,double& test_error,
		 // additional parameters for steplength
		 int steplength_rule, double theta,
		 int projection_rule, double projection_constant,
		 int n_grad_average) {
  
  uint num_examples = Labels.size();

  double startTime;
  double endTime;
  double evaluateMTime;
  //  uint nviol=0;

  // Initialize of classification vector W and the auxiliary
  // ("averaged") vector Wtil
  WeightVector W(dimension);
  WeightVector Wtil(dimension);
  double M = 0.0, DX = 1.0/sqrt(lambda);
  //double cycle_length = max_iter/100; // restart every cycle_length iterations
  //double avgScale = (num_iter_to_avg > max_iter) ? max_iter : num_iter_to_avg;
  
  //double *grad_values = (double*)malloc(n_grad_average*sizeof(double));
  //uint gidx = 0;
  //double grad_sum = 0;
  //double *wsnorm_values = (double*)malloc(n_grad_average*sizeof(double));
  //double wsnorm_sum = 0;

  // defines the iterate at which averaging should start
  int num_iter_to_avg;
  if (frac_iter_to_avg >= 1.0) 
    num_iter_to_avg = max_iter;
  else
    num_iter_to_avg = int(floor(frac_iter_to_avg * max_iter));

  cout << "\n in LearnRobust with lambda = " << lambda 
	    << ", theta = " << theta << endl
	    << "max_iter = " << max_iter  
	    << "  start averaging at iteration " << num_iter_to_avg << endl
	    << " stepsize rule = " << steplength_rule
	    << ", projection rule = " << projection_rule
	    << endl;
  
  // specify print frequency
  int printFreq = PRINTFREQ;

  int M_samples=0;
  train_time = 0;
  startTime = get_runtime();

  evaluateMTime = get_runtime();

  if (steplength_rule != 2) {
    // Preliminary calculations of constants used in steplength
    // settings. Base DX on the radius of the ball that w is restricted
    // to in the pegasos paper. Use sampling of the dataset to get the
    // bound on the square of the subgradient.
    DX = 1.0/sqrt(lambda);
    M_samples = num_examples / 100;

    if (M_samples <= 10 ) M_samples = 10;
    if (M_samples >= 100) M_samples = 100;
    M = 0.0;
    // average gradient contributions from the summation part
    for (int j=0; j < M_samples; j++) {
      // choose random example
      uint r = ((int)rand()) % num_examples;
      M += sqrt(Dataset[r].snorm());
    }
    // scale, and ad dthe other ingredients to make M an upper bound
    // estimate for E(||g||)
    M /= M_samples; M = M+1;
    cout << " LearnRobust: M = " << M << " DX = " << DX << endl;
  } else {
    // M and DX are not needed for the 1/t steplength rule; just set
    // them to zero
    M = 0.0; DX = 0.0;
  }

  evaluateMTime = get_runtime() - evaluateMTime;
  cout << "Time to evaluate M: " << evaluateMTime << endl;

  double sum_eta=0.0, sum_eta_new=0.0;
  //double eta; //, etamax, etamin;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // step length
    double eta, etamax;

    // step length
    if (steplength_rule == 0) { 
      // "1/sqrt(i)" rule from Nemirovski Sec 2.
      eta = theta * DX / (M * sqrt(1.0*(i+1))); 
      // bound the steplength by... something
      etamax = DX / (M * sqrt(1.0*(max_iter/10))); 
      if (eta > etamax) 
	eta = etamax;
    } else if (steplength_rule == 1) { 
      // "constant steplength" rule from Nemirovski Sec 2.2, scaled by
      // theta
      eta = theta * DX / (M * sqrt(1.0*max_iter));
      // solve numerical problems
      //if (projection_rule != 2)
    } else if (steplength_rule == 2) { 
      // theta/(lambda*i) steplength. Note that theta=1 appears ideal;
      // performance degrades rapidly away from this value, as
      // expected from the analysis in Nemirovski Sec 2.1. No need for
      // M or DX here.
      eta = theta / (lambda*(i+1));
    } else {
      cerr << "steplength_rule unrecognized - TERMINATE" << endl;
      exit(1);
    }

    // fix the scale parameters in W and Wtil to 1, if they fall below
    // MAKE_MY_A_ONE_THRES (see WeightVector.h)
    W.make_my_a_one();
    Wtil.make_my_a_one();

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;
    vector<double> grad_predictions;

    // calc a subgradient
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r = ((int)rand()) % num_examples;
      //      double r_where = (1.0 * r) / (1.0 * num_examples);
      //      if (i < 10) {
      //	cout << "i= " << i << "  r_where= " << r_where << endl;
      //      }

      // calculate prediction
      double prediction = W*Dataset[r];

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	grad_predictions.push_back(1-cur_loss);
      }
    }


    if (steplength_rule != 2) {
      // If we are averaging, update the "averaged" sequence using the
      // latest eta
      if (i+1 >= num_iter_to_avg)
        {
	  //Wtil.add(W, 1.0/avgScale);
	  sum_eta_new = sum_eta + eta;
	  double alpha;
	  alpha = sum_eta / sum_eta_new;
	  //      cout << "iteration " << i 
	  //		<< " eta=" << eta 
	  //		<< " alpha=" << alpha 
	  //		<< endl;
	  Wtil.scale(alpha);
	  Wtil.add(W,1.0-alpha);
	  sum_eta = sum_eta_new;
	}
    }

    // scale w 
    W.scale(1.0-eta*lambda);

    // and add subgradients
    //    if (i < 10) {
    //      cout << "size of grad_index: " << grad_index.size() << endl;
    //    }

    for (uint j=0; j<grad_index.size(); ++j) {
      W.add(Dataset[grad_index[j]],grad_weights[j]);
    }

    // and project onto the ball |w|_2 <= 1/sqrt(lambda) (if
    // projection_rule=0) or onto the ball of radius
    // projection_constant (if projection_rule=1) or else do no
    // projection (otherwise)
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection

    // disable this stuff (Sang's refined estimation of M)

    /*
    // M estimation by averaging norm^2 of gradient for a moving window
    if( i >= n_grad_average ) {
      grad_sum -= grad_values[gidx];
      wsnorm_sum -= wsnorm_values[gidx];
    }

    if( grad_index.size() > 0 ) {
      uint r = ((int)rand()) % grad_index.size();
      grad_values[gidx] = lambda*lambda*W.snorm() 
	-2*lambda*grad_predictions[r] + Dataset[grad_index[r]].snorm();
    } else {
      grad_values[gidx] = lambda*lambda*W.snorm();
    }
    grad_sum += grad_values[gidx];

    wsnorm_values[gidx] = W.snorm();
    wsnorm_sum += wsnorm_values[gidx];

    gidx = (gidx+1) % n_grad_average;

    DX = sqrt(wsnorm_sum/MIN(i+1,n_grad_average));
    M = sqrt(grad_sum/MIN(i+1,n_grad_average));
    */

    // Calculate and print objective value and error rates
    // periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      if (steplength_rule==2) 
	Wtil = W;
      obj_value = Wtil.snorm() * lambda / 2.0;
      loss_value = 0.0; train_error = 0.0;
      double cur_loss = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
	cur_loss = 1 - Labels[l]*(Wtil * Dataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) train_error += 1.0;
      }
      loss_value /= num_examples; train_error /= num_examples;
      obj_value += loss_value;

      // Calculate test_loss and test_error
      test_loss = 0.0; test_error = 0.0;
      for (uint l=0; l < testDataset.size(); ++l) {
	double cur_loss = 1 - testLabels[l]*(Wtil * testDataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	test_loss += cur_loss;
	if (cur_loss >= 1.0) test_error += 1.0;
      }
      if (testDataset.size() != 0) {
	test_loss /= testDataset.size();
	test_error /= testDataset.size();
      }

      cout << " iter " << i 
		<< " |Wtil|=" << sqrt(Wtil.snorm())
		<< " |W|=" << sqrt(W.snorm())
		<< " last eta=" << eta 
		<< " obj=" << obj_value 
		<< " trainErr=" << train_error
		<< " testErr=" << test_error
		<< " time=" << train_time
	//	        << " M=" << M
	//	        << " DX=" << DX
		<< endl;
      
      startTime = get_runtime();
    }
  }

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
  if (steplength_rule==2) {
    Wtil = W;
  }
  norm_value = sqrt(Wtil.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  train_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double cur_loss = 1 - Labels[i]*(Wtil * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) train_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  train_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;
  cout << "time to calculate the full objective (once): " 
	    << calc_obj_time << endl;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double cur_loss = 1 - testLabels[i]*(Wtil * testDataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
  

  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    Wtil.print(model_file);
    model_file.close();
  }

  // print out the time spent in "make_my_a_one()"
  cout << "make_my_a_one time: "
            << "  W    = " << W.make_my_a_one_time 
            << "  Wtil = " << Wtil.make_my_a_one_time << endl;

}


// LearnRobustAug is the same as LearnRobust, but it allows an
// intercept in the hyperplane. Hence, uses a different definition of
// the weight vector and the various operations performed on it.

// coded by SJW 9/14/09 and after.

void LearnRobustAug(// Inputs
		 vector<simple_sparse_vector>& Dataset,
		 vector<int>& Labels,
		 uint dimension,
		 vector<simple_sparse_vector>& testDataset,
		 vector<int>& testLabels,
		 double lambda, int max_iter,
		 int exam_per_iter, double frac_iter_to_avg,
		 string& model_filename,
		 // Outputs
		 double& train_time,double& calc_obj_time,
		 double& obj_value,double& norm_value,
		 double& loss_value,double& train_error,
		 double& test_loss,double& test_error,
		 // additional parameters for steplength
		 int steplength_rule, double theta,
		 int projection_rule, double projection_constant,
		 int n_grad_average) {
  
  uint num_examples = Labels.size();

  double startTime;
  double endTime;
  double evaluateMTime;
  //  uint nviol=0;

  // Initialize of classification vector W and the auxiliary
  // ("averaged") vector Wtil
  AugWeightVector W(dimension);
  AugWeightVector Wtil(dimension);
  double M = 0.0, DX = 1.0/sqrt(lambda);
  //double cycle_length = max_iter/100; // restart every cycle_length iterations
  //double avgScale = (num_iter_to_avg > max_iter) ? max_iter : num_iter_to_avg;
  
  //double *grad_values = (double*)malloc(n_grad_average*sizeof(double));
  //uint gidx = 0;
  //double grad_sum = 0;
  //double *wsnorm_values = (double*)malloc(n_grad_average*sizeof(double));
  //double wsnorm_sum = 0;

  // defines the iterate at which averaging should start
  int num_iter_to_avg;
  if (frac_iter_to_avg >= 1.0) 
    num_iter_to_avg = max_iter;
  else
    num_iter_to_avg = int(floor(frac_iter_to_avg * max_iter));

  cout << "\n in LearnRobustAug with lambda = " << lambda 
	    << ", theta = " << theta << endl
	    << "max_iter = " << max_iter  
	    << "  start averaging at iteration " << num_iter_to_avg << endl
	    << " stepsize rule = " << steplength_rule
	    << ", projection rule = " << projection_rule
	    << endl;
  
  // specify print frequency
  int printFreq = PRINTFREQ;

  int M_samples=0;
  train_time = 0;
  startTime = get_runtime();

  evaluateMTime = get_runtime();

  if (steplength_rule != 2) {
    // Preliminary calculations of constants used in steplength
    // settings. Base DX on the radius of the ball that w is restricted
    // to in the pegasos paper. Use sampling of the dataset to get the
    // bound on the square of the subgradient.
    DX = 1.0/sqrt(lambda);
    M_samples = num_examples / 100;

    if (M_samples <= 10 ) M_samples = 10;
    if (M_samples >= 100) M_samples = 100;
    M = 0.0;
    // average gradient contributions from the summation part
    for (int j=0; j < M_samples; j++) {
      // choose random example
      uint r = ((int)rand()) % num_examples;
      // add 1 to the squared norm of x_r, to account for the appended +1
      M += sqrt(Dataset[r].snorm()+1.0);
    }
    // scale, and add the other ingredients to make M an upper bound
    // estimate for E(||g||)
    M /= M_samples; M = M+1;
    cout << " LearnRobust: M = " << M << " DX = " << DX << endl;
  } else {
    // M and DX are not needed for the 1/t steplength rule; just set
    // them to zero
    M = 0.0; DX = 0.0;
  }

  evaluateMTime = get_runtime() - evaluateMTime;
  cout << "Time to evaluate M: " << evaluateMTime << endl;

  double sum_eta=0.0, sum_eta_new=0.0;
  //double eta; //, etamax, etamin;

  // ---------------- Main Loop -------------------
  for (int i = 0; i < max_iter; ++i) {

    // step length
    double eta, etamax;

    // step length
    if (steplength_rule == 0) { 
      // "1/sqrt(i)" rule from Nemirovski Sec 2.
      eta = theta * DX / (M * sqrt(1.0*(i+1))); 
      // bound the steplength by... something
      etamax = DX / (M * sqrt(1.0*(max_iter/10))); 
      if (eta > etamax) 
	eta = etamax;
    } else if (steplength_rule == 1) { 
      // "constant steplength" rule from Nemirovski Sec 2.2, scaled by
      // theta
      eta = theta * DX / (M * sqrt(1.0*max_iter));
      // solve numerical problems
      //if (projection_rule != 2)
    } else if (steplength_rule == 2) { 
      // theta/(lambda*i) steplength. Note that theta=1 appears ideal;
      // performance degrades rapidly away from this value, as
      // expected from the analysis in Nemirovski Sec 2.1. No need for
      // M or DX here.
      eta = theta / (lambda*(i+1));
    } else {
      cerr << "steplength_rule unrecognized - TERMINATE" << endl;
      exit(1);
    }

    // fix the scale parameters in W and Wtil to 1, if they fall below
    // MAKE_MY_A_ONE_THRES (see WeightVector.h)
    W.make_my_a_one();
    Wtil.make_my_a_one();

    // gradient indices and losses
    vector<uint> grad_index;
    vector<double> grad_weights;
    vector<double> grad_predictions;

    // calc a subgradient
    for (int j=0; j < exam_per_iter; ++j) {

      // choose random example
      uint r = ((int)rand()) % num_examples;
      //      double r_where = (1.0 * r) / (1.0 * num_examples);
      //      if (i < 10) {
      //	cout << "i= " << i << "  r_where= " << r_where << endl;
      //      }

      // calculate prediction
      double prediction = W*Dataset[r];

      // calculate loss
      double cur_loss = 1 - Labels[r]*prediction;
      if (cur_loss < 0.0) cur_loss = 0.0;

      // and add to the gradient
      if (cur_loss > 0.0) {
	grad_index.push_back(r);
	grad_weights.push_back(eta*Labels[r]/exam_per_iter);
	grad_predictions.push_back(1-cur_loss);
      }
    }


    if (steplength_rule != 2) {
      // If we are averaging, update the "averaged" sequence using the
      // latest eta
      if (i+1 >= num_iter_to_avg)
        {
	  //Wtil.add(W, 1.0/avgScale);
	  sum_eta_new = sum_eta + eta;
	  double alpha;
	  alpha = sum_eta / sum_eta_new;
	  //      cout << "iteration " << i 
	  //		<< " eta=" << eta 
	  //		<< " alpha=" << alpha 
	  //		<< endl;
	  Wtil.scale(alpha);
	  Wtil.add(W,1.0-alpha);
	  sum_eta = sum_eta_new;
	}
    }

    // scale w 
    W.scaleW(1.0-eta*lambda);

    // and add subgradients
    //    if (i < 10) {
    //      cout << "size of grad_index: " << grad_index.size() << endl;
    //    }

    for (uint j=0; j<grad_index.size(); ++j) {
      W.add(Dataset[grad_index[j]],grad_weights[j]);
    }

    // and project onto the ball |w|_2 <= 1/sqrt(lambda) (if
    // projection_rule=0) or onto the ball of radius
    // projection_constant (if projection_rule=1) or else do no
    // projection (otherwise)
    if (projection_rule == 0) { // Pegasos projection rule
      double norm2 = W.snorm();
      if (norm2 > 1.0/lambda) {
	W.scale(sqrt(1.0/(lambda*norm2)));
      }
    } else if (projection_rule == 1) { // other projection
      double norm2 = W.snorm();
      if (norm2 > (projection_constant*projection_constant)) {
	W.scale(projection_constant/sqrt(norm2));
      }
    } // else -- no projection


    // Calculate and print objective value and error rates
    // periodically
    if (i%printFreq == 0) {
      train_time += (get_runtime() - startTime);

      if (steplength_rule==2) 
	Wtil = W;
      obj_value = Wtil.snorm() * lambda / 2.0;
      loss_value = 0.0; train_error = 0.0;
      double cur_loss = 0.0;
      for (uint l=0; l < Dataset.size(); ++l) {
	cur_loss = 1 - Labels[l]*(Wtil * Dataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	loss_value += cur_loss;
	if (cur_loss >= 1.0) train_error += 1.0;
      }
      loss_value /= num_examples; train_error /= num_examples;
      obj_value += loss_value;

      // Calculate test_loss and test_error
      test_loss = 0.0; test_error = 0.0;
      for (uint l=0; l < testDataset.size(); ++l) {
	double cur_loss = 1 - testLabels[l]*(Wtil * testDataset[l]); 
	if (cur_loss < 0.0) cur_loss = 0.0;
	test_loss += cur_loss;
	if (cur_loss >= 1.0) test_error += 1.0;
      }
      if (testDataset.size() != 0) {
	test_loss /= testDataset.size();
	test_error /= testDataset.size();
      }

      cout << " iter " << i 
		<< " |Wtil|=" << sqrt(Wtil.snorm())
		<< " Wtil.beta=" << Wtil.beta()
		<< " |W|=" << sqrt(W.snorm())
		<< " W.beta=" << W.beta()
		<< " last eta=" << eta 
		<< " obj=" << obj_value 
		<< " trainErr=" << train_error
		<< " testErr=" << test_error
		<< " time=" << train_time
	//	        << " M=" << M
	//	        << " DX=" << DX
		<< endl;
      
      startTime = get_runtime();
    }
  }

  // update timeline
  endTime = get_runtime();
  train_time += (endTime - startTime);
  startTime = get_runtime();

  // Calculate objective value
  if (steplength_rule==2) {
    Wtil = W;
  }
  norm_value = sqrt(Wtil.snorm());
  obj_value = norm_value * norm_value * lambda / 2.0;
  loss_value = 0.0;
  train_error = 0.0;
  for (uint i=0; i < Dataset.size(); ++i) {
    double cur_loss = 1 - Labels[i]*(Wtil * Dataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    loss_value += cur_loss;
    if (cur_loss >= 1.0) train_error += 1.0;
  }
  loss_value /= num_examples;
  obj_value += loss_value;
  train_error /= num_examples;

  endTime = get_runtime();
  calc_obj_time = endTime - startTime;
  cout << "time to calculate the full objective (once): " 
	    << calc_obj_time << endl;

  // Calculate test_loss and test_error
  test_loss = 0.0;
  test_error = 0.0;
  for (uint i=0; i < testDataset.size(); ++i) {
    double cur_loss = 1 - testLabels[i]*(Wtil * testDataset[i]); 
    if (cur_loss < 0.0) cur_loss = 0.0;
    test_loss += cur_loss;
    if (cur_loss >= 1.0) test_error += 1.0;
  }
  if (testDataset.size() != 0) {
    test_loss /= testDataset.size();
    test_error /= testDataset.size();
  }
  

  // finally, print the model to the model_file
  if (model_filename != "noModelFile") {
    ofstream model_file(model_filename.c_str());
    if (!model_file.good()) {
      cerr << "error w/ " << model_filename << endl;
      exit(EXIT_FAILURE);
    }
    Wtil.print(model_file);
    model_file.close();
  }

  // print out the time spent in "make_my_a_one()"
  cout << "make_my_a_one time: "
            << "  W    = " << W.make_my_a_one_time 
            << "  Wtil = " << Wtil.make_my_a_one_time << endl;

}

// ------------------------------------------------------------//
// ---------------- READING DATA ------------------------------//
// ------------------------------------------------------------//
void ReadData(// input
	      string& data_filename,
	      // output
	      vector<simple_sparse_vector> & Dataset,
	      vector<int> & Labels,
	      uint& dimension,
	      double& sparsity,
	      double& readingTime) {
  
  dimension = 0;

  // Start a timer
  double startTime = get_runtime();

  // OPEN DATA FILE
  // =========================
  ifstream data_file(data_filename.c_str());
  if (!data_file.good()) {
    cerr << "error w/ " << data_filename << endl;
    exit(EXIT_FAILURE);
  }

  
  // Read SVM-Light data file
  // ========================
  int num_examples = 0, total_features = 0;
  string buf;
  while (getline(data_file,buf)) {
    // ignore lines which begin with #
    if (buf[0] == '#') continue;
    // Erase what comes after #
    size_t pos = buf.find('#');
    if (pos < buf.size()) {
      buf.erase(pos);
    }
    // replace ':' with white space
    int n=0;
    for (size_t pos=0; pos < buf.size(); ++pos)
      if (buf[pos] == ':') {
	n++; buf[pos] = ' ';
      }
    total_features += n;
    // read from the string
    istringstream is(buf);
    int label = 0;
    is >> label;
    if (label != 1 && label != -1) {
      cerr << "Error reading SVM-light format. Abort." << endl;
      exit(EXIT_FAILURE);
    }
    Labels.push_back(label);
    simple_sparse_vector instance(is,n);
    Dataset.push_back(instance);
    num_examples++;
    uint cur_max_ind = instance.max_index() + 1;
    if (cur_max_ind > dimension) dimension = cur_max_ind;
  }

  sparsity = (1.0 * total_features) / (num_examples * dimension);

  data_file.close();


#ifdef nodef
  cerr << "num_examples = " << num_examples 
	    << " dimension = " << dimension
	    << " Dataset.size = " << Dataset.size() 
	    << " Labels.size = " << Labels.size() 
	    << " Sparsity = " << sparsity
	    << endl;
#endif
    
  
  // update timeline
  readingTime = get_runtime() - startTime;
  
}


// -------------------------------------------------------------//
// ---------------------- Experiments mode ---------------------//
// -------------------------------------------------------------//

  class ExperimentStruct {
  public:
    ExperimentStruct() { }
    void Load(istringstream& is) {
      is >> lambda >> max_iter >> exam_per_iter >> num_iter_to_avg
	 >> eta_rule >> eta_constant >> projection_rule 
	 >> projection_constant;
    }
    void Print() {
      cout << lambda << "\t\t" << max_iter << "\t\t" << exam_per_iter << "\t\t" 
		<< num_iter_to_avg << "\t\t" << eta_rule << "\t\t" << eta_constant 
		<< "\t\t" << projection_rule << "\t\t" << projection_constant << "\t\t";
    }
    void PrintHead() {
      cerr << "lambda\t\tT\t\tk\t\tnumValid\t\te_r\t\te_c\t\tp_r\t\tp_c\t\t";
    }
    double lambda, eta_constant, projection_constant;
    uint max_iter,exam_per_iter,num_iter_to_avg;
    int eta_rule,projection_rule;
  };

  class ResultStruct {
  public:
    ResultStruct() : trainTime(0), calc_obj_time(0),
		     norm_value(0.0),loss_value(0.0),
		     zero_one_error(0.0),obj_value(0.0),
		     test_loss(0.0), test_error(0.0) { }
    void Print() {
      cout << trainTime << "\t\t" 
		<< calc_obj_time << "\t\t" 
		<< norm_value  << "\t\t" 
		<< loss_value << "\t\t"  
		<< zero_one_error  << "\t\t" 
		<< obj_value << "\t\t"
		<< test_loss << "\t\t"
		<< test_error << "\t\t";

    }
    void PrintHead() {
      cerr << "tTime\t\tcoTime\t\t||w||\t\tL\t\tL0-1\t\tobj_value\t\ttest_L\t\ttestL0-1\t\t";
    }

    double trainTime, calc_obj_time;
    double norm_value,loss_value,zero_one_error,obj_value,test_loss,test_error;
  };


void run_experiments(string& experiments_filename,
		     vector<simple_sparse_vector>& Dataset,
		     vector<int>& Labels,
		     uint dimension,
		     vector<simple_sparse_vector>& testDataset,
		     vector<int>& testLabels) {

  // open the experiments file
  ifstream exp_file(experiments_filename.c_str());
  if (!exp_file.good()) {
    cerr << "error w/ " << experiments_filename << endl;
    exit(EXIT_FAILURE);
  }

  // read the entire experiment specification

  
  uint num_experiments = 0;
  vector<ExperimentStruct> v;
  string buf;
  while (getline(exp_file,buf)) {
    // read from the string
    istringstream is(buf);
    ExperimentStruct tmp; tmp.Load(is);
    v.push_back(tmp);
    num_experiments++;
  }
  exp_file.close();

  // run all the experiments
  vector<ResultStruct> res(num_experiments);
  string lala = "noModelFile";
  for (uint i=0; i<num_experiments; ++i) {
    LearnReturnLast(Dataset,Labels,dimension,testDataset,testLabels,
		     v[i].lambda,v[i].max_iter,v[i].exam_per_iter,
		     //v[i].num_iter_to_avg,
		     lala,
		     res[i].trainTime,res[i].calc_obj_time,res[i].obj_value,
		     res[i].norm_value,
		     res[i].loss_value,res[i].zero_one_error,
		     res[i].test_loss,res[i].test_error,
		     v[i].eta_rule,v[i].eta_constant,
		     v[i].projection_rule,v[i].projection_constant);
  }

  // print results
  v[0].PrintHead(); res[0].PrintHead(); 
  cout << endl;
  for (uint i=0; i<num_experiments; ++i) {
    v[i].Print();
    res[i].Print();
    cout << endl;
  }

  cerr << endl;

}
